@@ -1,16 +1,13 @@
 {----------------------------------------------------------------------------
-      ___    ___   ___    ___   __________   __________
-    /  /   /  /  /  /   /  /  /  _______/  /  _______/       Hugs 1.4
-   /  /___/  /  /  /   /  /  /  / _____   /  /______
-  /  ____   /  /  /   /  /  /  / /_   /  /______   /  The Nottingham and Yale
- /  /   /  /  /  /___/  /  /  /___/  /  _______/  /    Haskell User's System
-/__/   /__/  /_________/  /_________/  /_________/
+__   __ __  __  ____   ___    _______________________________________________
+||   || ||  || ||  || ||__    Hugs 98: The Nottingham and Yale Haskell system
+||___|| ||__|| ||__||  __||   Copyright (c) 1994-1999
+||---||         ___||         World Wide Web: http://haskell.org/hugs
+||   ||                       Report bugs to: hugs-bugs@haskell.org
+||   || Version: February 1999_______________________________________________
 
-   Copyright (c) The University of Nottingham and Yale University, 1994-1997.
-    Bug reports: hugs-bugs@haskell.org.   Web: http://www.haskell.org/hugs.
-
- This is the Hugs Standard Prelude, based very closely on the Standard
- Prelude for Haskell 1.4.
+ This is the Hugs 98 Standard Prelude, based very closely on the Standard
+ Prelude for Haskell 98.
 
  WARNING: This file is an integral part of the Hugs source code.  Changes to
  the definitions in this file without corresponding modifications in other
@@ -18,12 +15,16 @@
  normal circumstances, you should not attempt to modify this file in any way!
 
 -----------------------------------------------------------------------------
-     Hugs is subject to conditions of use and distribution; see the file
-      "NOTICE" included with the main distribution for further details.
+ The Hugs 98 system is Copyright (c) Mark P Jones, Alastair Reid, the
+ Yale Haskell Group, and the Oregon Graduate Institute of Science and
+ Technology, 1994-1999, All rights reserved.  It is distributed as
+ free software under the license in the file "License", which is
+ included in the distribution.
 ----------------------------------------------------------------------------}
 
 module Prelude (
 --  module PreludeList,
+    map, (++), concat, filter,
     head, last, tail, init, null, length, (!!),
     foldl, foldl1, scanl, scanl1, foldr, foldr1, scanr, scanr1,
     iterate, repeat, replicate, cycle,
@@ -35,11 +36,11 @@ module Prelude (
 --  module PreludeText, 
     ReadS, ShowS,
     Read(readsPrec, readList),
-    Show(showsPrec, showList),
-    reads, shows, show, read, lex,
+    Show(show, showsPrec, showList),
+    reads, shows, read, lex,
     showChar, showString, readParen, showParen,
 --  module PreludeIO,
-    FilePath, IOError, fail, userError, catch,
+    FilePath, IOError, ioError, userError, catch,
     putChar, putStr, putStrLn, print,
     getChar, getLine, getContents, interact,
     readFile, writeFile, appendFile, readIO, readLn,
@@ -47,10 +48,11 @@ module Prelude (
     Ix(range, index, inRange, rangeSize),
 --  module Char,
     isAscii, isControl, isPrint, isSpace, isUpper, isLower,
-    isAlpha, isDigit, isOctDigit, isHexDigit, isAlphanum,
+    isAlpha, isDigit, isOctDigit, isHexDigit, isAlphaNum,
     digitToInt, intToDigit,
     toUpper, toLower,
     ord, chr,
+    readLitChar, showLitChar, lexLitChar,
 --  module Numeric
     showSigned, showInt,
     readSigned, readInt,
@@ -59,13 +61,15 @@ module Prelude (
 --  module Ratio,
     Ratio, Rational, (%), numerator, denominator, approxRational,
 --  Non-standard exports
-    ST, IO(..), IOResult(..), primExitWith,
+    IO(..), IOResult(..), primExitWith, Addr, Word, StablePtr, ForeignObj,
+    basicIORun, IOFinished(..),
+    threadToIOResult,
 
     Bool(False, True),
     Maybe(Nothing, Just),
     Either(Left, Right),
     Ordering(LT, EQ, GT),
-    Char, String, Int, Integer, Float, Double, IO, Void,
+    Char, String, Int, Integer, Float, Double, IO,
 --  List type: []((:), [])
     (:),
 --  Tuple types: (,), (,,), etc.
@@ -74,10 +78,9 @@ module Prelude (
     Rec, EmptyRec, EmptyRow, -- non-standard, should only be exported if TREX
     Eq((==), (/=)),
     Ord(compare, (<), (<=), (>=), (>), max, min),
-    Enum(toEnum, fromEnum, enumFrom, enumFromThen,
+    Enum(succ, pred, toEnum, fromEnum, enumFrom, enumFromThen,
          enumFromTo, enumFromThenTo),
     Bounded(minBound, maxBound),
-    Eval(seq, strict),
 --  Num((+), (-), (*), negate, abs, signum, fromInteger),
     Num((+), (-), (*), negate, abs, signum, fromInteger, fromInt),
     Real(toRational),
@@ -90,19 +93,18 @@ module Prelude (
     RealFrac(properFraction, truncate, round, ceiling, floor),
     RealFloat(floatRadix, floatDigits, floatRange, decodeFloat,
               encodeFloat, exponent, significand, scaleFloat, isNaN,
-              isInfinite, isDenormalized, isIEEE, isNegativeZero),
-    Monad((>>=), (>>), return),
-    MonadZero(zero),
-    MonadPlus((++)),
-    Functor(map),
-    succ, pred,
-    mapM, mapM_, guard, accumulate, sequence, filter, concat, applyM,
+              isInfinite, isDenormalized, isIEEE, isNegativeZero, atan2),
+    Monad((>>=), (>>), return, fail),
+    Functor(fmap),
+    mapM, mapM_, sequence, sequence_, (=<<),
     maybe, either,
     (&&), (||), not, otherwise,
     subtract, even, odd, gcd, lcm, (^), (^^), 
-    fromIntegral, fromRealFrac, atan2,
+    fromIntegral, realToFrac,
     fst, snd, curry, uncurry, id, const, (.), flip, ($), until,
-    asTypeOf, error, undefined ) where
+    asTypeOf, error, undefined,
+    seq, ($!)
+  ) where
 
 -- Standard value bindings {Prelude} ----------------------------------------
 
@@ -111,17 +113,22 @@ infixl 9  !!
 infixr 8  ^, ^^, **
 infixl 7  *, /, `quot`, `rem`, `div`, `mod`, :%, %
 infixl 6  +, -
-infixr 5  :, ++
+--infixr 5  :    -- this fixity declaration is hard-wired into Hugs
+infixr 5  ++
 infix  4  ==, /=, <, <=, >=, >, `elem`, `notElem`
 infixr 3  &&
 infixr 2  ||
 infixl 1  >>, >>=
-infixr 0  $, `seq`
+infixr 1  =<<
+infixr 0  $, $!, `seq`
 
 -- Equality and Ordered classes ---------------------------------------------
 
 class Eq a where
     (==), (/=) :: a -> a -> Bool
+
+    -- Minimal complete definition: (==) or (/=)
+    x == y      = not (x/=y)
     x /= y      = not (x==y)
 
 class (Eq a) => Ord a where
@@ -129,6 +136,8 @@ class (Eq a) => Ord a where
     (<), (<=), (>=), (>)   :: a -> a -> Bool
     max, min               :: a -> a -> a
 
+    -- Minimal complete definition: (<=) or compare
+    -- using compare can be more efficient for complex types
     compare x y | x==y      = EQ
 		| x<=y      = LT
 		| otherwise = GT
@@ -145,18 +154,21 @@ class (Eq a) => Ord a where
 
 class Bounded a where
     minBound, maxBound :: a
+    -- Minimal complete definition: All
 
 -- Numeric classes ----------------------------------------------------------
 
-class (Eq a, Show a, Eval a) => Num a where
+class (Eq a, Show a) => Num a where
     (+), (-), (*)  :: a -> a -> a
     negate         :: a -> a
     abs, signum    :: a -> a
     fromInteger    :: Integer -> a
     fromInt        :: Int -> a
 
+    -- Minimal complete definition: All, except negate or (-)
     x - y           = x + negate y
     fromInt         = fromIntegral
+    negate x        = 0 - x
 
 class (Num a, Ord a) => Real a where
     toRational     :: a -> Rational
@@ -168,6 +180,7 @@ class (Real a, Enum a) => Integral a where
     toInteger           :: a -> Integer
     toInt               :: a -> Int
 
+    -- Minimal complete definition: quotRem and toInteger
     n `quot` d           = q where (q,r) = quotRem n d
     n `rem` d            = r where (q,r) = quotRem n d
     n `div` d            = q where (q,r) = divMod n d
@@ -184,8 +197,10 @@ class (Num a) => Fractional a where
     fromRational :: Rational -> a
     fromDouble   :: Double -> a
 
+    -- Minimal complete definition: fromRational and ((/) or recip)
     recip x       = 1 / x
     fromDouble    = fromRational . toRational
+    x / y         = x * recip y
 
 
 class (Fractional a) => Floating a where
@@ -197,6 +212,9 @@ class (Fractional a) => Floating a where
     sinh, cosh, tanh    :: a -> a
     asinh, acosh, atanh :: a -> a
 
+    -- Minimal complete definition: pi, exp, log, sin, cos, sinh, cosh,
+    --				    asinh, acosh, atanh
+    pi                   = 4 * atan 1
     x ** y               = exp (log x * y)
     logBase x y          = log y / log x
     sqrt x               = x ** 0.5
@@ -213,6 +231,7 @@ class (Real a, Fractional a) => RealFrac a where
     truncate, round  :: (Integral b) => a -> b
     ceiling, floor   :: (Integral b) => a -> b
 
+    -- Minimal complete definition: properFraction
     truncate x        = m where (m,_) = properFraction x
 
     round x           = let (n,r) = properFraction x
@@ -239,13 +258,28 @@ class (RealFrac a, Floating a) => RealFloat a where
     scaleFloat       :: Int -> a -> a
     isNaN, isInfinite, isDenormalized, isNegativeZero, isIEEE
 		     :: a -> Bool
+    atan2	     :: a -> a -> a
 
+    -- Minimal complete definition: All, except exponent, signficand,
+    --				    scaleFloat, atan2
     exponent x        = if m==0 then 0 else n + floatDigits x
 			where (m,n) = decodeFloat x
     significand x     = encodeFloat m (- floatDigits x)
 			where (m,_) = decodeFloat x
     scaleFloat k x    = encodeFloat m (n+k)
 			where (m,n) = decodeFloat x
+    atan2 y x
+      | x>0           = atan (y/x)
+      | x==0 && y>0   = pi/2
+      | x<0 && y>0    = pi + atan (y/x)
+      | (x<=0 && y<0) ||
+        (x<0 && isNegativeZero y) ||
+        (isNegativeZero x && isNegativeZero y)
+		      = - atan2 (-y) x
+      | y==0 && (x<0 || isNegativeZero x)
+		      = pi    -- must be after the previous test on zero y
+      | x==0 && y==0  = y     -- must be after the other double zero tests
+      | otherwise     = x + y -- x or y is a NaN, return a NaN (via +)
 
 -- Numeric functions --------------------------------------------------------
 
@@ -278,18 +312,8 @@ x ^^ n          = if n >= 0 then x ^ n else recip (x^(-n))
 fromIntegral   :: (Integral a, Num b) => a -> b
 fromIntegral    = fromInteger . toInteger
 
-fromRealFrac   :: (RealFrac a, Fractional b) => a -> b
-fromRealFrac    = fromRational . toRational
-
-atan2          :: (RealFloat a) => a -> a -> a
-atan2 y x       = case (signum y, signum x) of
-		    ( 0, 1) ->  0
-		    ( 1, 0) ->  pi/2
-		    ( 0,-1) ->  pi
-		    (-1, 0) -> -pi/2
-		    ( _, 1) -> atan (y/x)
-		    ( _,-1) -> atan (y/x) + pi
-		    ( 0, 0) -> error "Prelude.atan2: atan2 of origin"
+realToFrac     :: (Real a, Fractional b) => a -> b
+realToFrac      = fromRational . toRational
 
 -- Index and Enumeration classes --------------------------------------------
 
@@ -304,6 +328,7 @@ class (Ord a) => Ix a where
              | otherwise  = index r u + 1
 
 class Enum a where
+    succ, pred           :: a -> a
     toEnum               :: Int -> a
     fromEnum             :: a -> Int
     enumFrom             :: a -> [a]              -- [n..]
@@ -311,13 +336,14 @@ class Enum a where
     enumFromTo           :: a -> a -> [a]         -- [n..m]
     enumFromThenTo       :: a -> a -> a -> [a]    -- [n,n'..m]
 
+    -- Minimal complete definition: toEnum, fromEnum
+    succ                  = toEnum . (1+)       . fromEnum
+    pred                  = toEnum . subtract 1 . fromEnum
+    enumFrom x            = map toEnum [ fromEnum x ..]
+    enumFromThen x y      = map toEnum [ fromEnum x, fromEnum y ..]
     enumFromTo x y        = map toEnum [ fromEnum x .. fromEnum y ]
     enumFromThenTo x y z  = map toEnum [ fromEnum x, fromEnum y .. fromEnum z ]
 
-succ, pred :: Enum a => a -> a
-succ        = toEnum . (1+)       . fromEnum
-pred        = toEnum . subtract 1 . fromEnum
-
 -- Read and Show classes ------------------------------------------------------
 
 type ReadS a = String -> [(a,String)]
@@ -326,6 +352,8 @@ type ShowS   = String -> String
 class Read a where
     readsPrec :: Int -> ReadS a
     readList  :: ReadS [a]
+
+    -- Minimal complete definition: readsPrec
     readList   = readParen False (\r -> [pr | ("[",s) <- lex r,
 					      pr      <- readl s ])
 		 where readl  s = [([],t)   | ("]",t) <- lex s] ++
@@ -337,68 +365,57 @@ class Read a where
 					      (xs,v)  <- readl' u]
 
 class Show a where
+    show      :: a -> String
     showsPrec :: Int -> a -> ShowS
     showList  :: [a] -> ShowS
 
+    -- Minimal complete definition: show or showsPrec
+    show x          = showsPrec 0 x ""
+    showsPrec _ x s = show x ++ s
     showList []     = showString "[]"
     showList (x:xs) = showChar '[' . shows x . showl xs
 		      where showl []     = showChar ']'
-			    showl (x:xs) = showString ", " . shows x . showl xs
+			    showl (x:xs) = showChar ',' . shows x . showl xs
 
 -- Monad classes ------------------------------------------------------------
 
 class Functor f where
-    map :: (a -> b) -> (f a -> f b)
+    fmap :: (a -> b) -> (f a -> f b)
 
 class Monad m where
-    (>>=)  :: m a -> (a -> m b) -> m b
     return :: a -> m a
-
+    (>>=)  :: m a -> (a -> m b) -> m b
     (>>)   :: m a -> m b -> m b
+    fail   :: String -> m a
+
+    -- Minimal complete definition: (>>=), return
     p >> q  = p >>= \ _ -> q
+    fail s  = error s
 
-class Monad m => MonadZero m where
-    zero   :: m a
+sequence       :: Monad m => [m a] -> m [a]
+sequence []     = return []
+sequence (c:cs) = do x  <- c
+		     xs <- sequence cs
+		     return (x:xs)
 
-class MonadZero m => MonadPlus m where
-    (++)   :: m a -> m a -> m a
-
-accumulate       :: Monad m => [m a] -> m [a]
-accumulate []     = return []
-accumulate (c:cs) = do x  <- c
-		       xs <- accumulate cs
-		       return (x:xs)
-
-sequence         :: Monad m => [m a] -> m ()
-sequence          = foldr (>>) (return ())
+sequence_        :: Monad m => [m a] -> m ()
+sequence_         = foldr (>>) (return ())
 
 mapM             :: Monad m => (a -> m b) -> [a] -> m [b]
-mapM f            = accumulate . map f
+mapM f            = sequence . map f
 
 mapM_            :: Monad m => (a -> m b) -> [a] -> m ()
-mapM_ f           = sequence . map f
+mapM_ f           = sequence_ . map f
 
-guard            :: MonadZero m => Bool -> m ()
-guard p           = if p then return () else zero
-
-filter           :: MonadZero m => (a -> Bool) -> m a -> m a
-filter p          = applyM (\x -> if p x then return x else zero)
-
-concat           :: MonadPlus m => [m a] -> m a
-concat            = foldr (++) zero
-
-applyM           :: Monad m => (a -> m b) -> m a -> m b
-applyM f x        = x >>= f
+(=<<)            :: Monad m => (a -> m b) -> m a -> m b
+f =<< x           = x >>= f
 
 -- Evaluation and strictness ------------------------------------------------
 
-class Eval a where
-    strict :: (a -> b) -> a -> b
-    seq    :: a -> b -> b
+primitive seq           :: a -> b -> b
 
--- Void type ----------------------------------------------------------------
-
-data Void      -- Void type has only one element, namely bottom.
+primitive ($!) "strict" :: (a -> b) -> a -> b
+-- f $! x                = x `seq` f x
 
 -- Trivial type -------------------------------------------------------------
 
@@ -482,7 +499,7 @@ instance Ix Char where
 instance Read Char where
     readsPrec p      = readParen False
 			    (\r -> [(c,t) | ('\'':s,t) <- lex r,
-					    (c,_)      <- readLitChar s ])
+					    (c,"\'")   <- readLitChar s ])
     readList = readParen False (\r -> [(l,t) | ('"':s, t) <- lex r,
 					       (l,_)      <- readl s ])
 	       where readl ('"':s)      = [("",s)]
@@ -503,7 +520,7 @@ instance Bounded Char where
     maxBound = '\255'
 
 isAscii, isControl, isPrint, isSpace            :: Char -> Bool
-isUpper, isLower, isAlpha, isDigit, isAlphanum  :: Char -> Bool
+isUpper, isLower, isAlpha, isDigit, isAlphaNum  :: Char -> Bool
 
 isAscii c              =  fromEnum c < 128
 isControl c            =  c < ' ' ||  c == '\DEL'
@@ -514,7 +531,7 @@ isUpper c              =  c >= 'A'   &&  c <= 'Z'
 isLower c              =  c >= 'a'   &&  c <= 'z'
 isAlpha c              =  isUpper c  ||  isLower c
 isDigit c              =  c >= '0'   &&  c <= '9'
-isAlphanum c           =  isAlpha c  ||  isDigit c
+isAlphaNum c           =  isAlpha c  ||  isDigit c
 
 -- Digit conversion operations
 digitToInt :: Char -> Int
@@ -553,20 +570,14 @@ maybe n f Nothing  = n
 maybe n f (Just x) = f x
 
 instance Functor Maybe where
-    map f Nothing  = Nothing
-    map f (Just x) = Just (f x)
+    fmap f Nothing  = Nothing
+    fmap f (Just x) = Just (f x)
 
 instance Monad Maybe where
-    Just x  >>= k  =  k x
-    Nothing >>= k  =  Nothing
-    return         =  Just
-
-instance MonadZero Maybe where
-    zero           = Nothing
-
-instance MonadPlus Maybe where
-    Nothing ++ ys  = ys
-    xs      ++ ys  = xs
+    Just x  >>= k = k x
+    Nothing >>= k = Nothing
+    return        = Just
+    fail s        = Nothing
 
 -- Either type --------------------------------------------------------------
 
@@ -598,20 +609,13 @@ instance Ord a => Ord [a] where
     compare (x:xs) (y:ys) = primCompAux x y (compare xs ys)
 
 instance Functor [] where
-    map f []     = []
-    map f (x:xs) = f x : map f xs
+    fmap = map
 
 instance Monad [ ] where
-    (x:xs) >>= f  =  f x ++ (xs >>= f)
-    []     >>= f  =  []
-    return x      =  [x]
-
-instance MonadZero [ ] where
-    zero = []
-
-instance MonadPlus [ ] where
-    []     ++ ys = ys
-    (x:xs) ++ ys = x : (xs ++ ys)
+    (x:xs) >>= f = f x ++ (xs >>= f)
+    []     >>= f = []
+    return x     = [x]
+    fail s       = []
 
 instance Read a => Read [a]  where
     readsPrec p = readList
@@ -624,14 +628,6 @@ instance Show a => Show [a]  where
 -- data (a,b) = (a,b) deriving (Eq, Ord, Ix, Read, Show)
 -- etc..
 
--- Functions ----------------------------------------------------------------
-
-instance Show (a -> b) where
-    showsPrec p f = showString "<<function>>"
-
-instance Functor ((->) a) where
-    map f g x = f (g x)
-
 -- Standard Integral types --------------------------------------------------
 
 data Int      -- builtin datatype of fixed size integers
@@ -758,11 +754,12 @@ numericEnumFrom        :: Real a => a -> [a]
 numericEnumFromThen    :: Real a => a -> a -> [a]
 numericEnumFromTo      :: Real a => a -> a -> [a]
 numericEnumFromThenTo  :: Real a => a -> a -> a -> [a]
-numericEnumFrom n            = n : strict numericEnumFrom (n+1)
+numericEnumFrom n            = n : (numericEnumFrom $! (n+1))
 numericEnumFromThen n m      = iterate ((m-n)+) n
 numericEnumFromTo n m        = takeWhile (<= m) (numericEnumFrom n)
-numericEnumFromThenTo n n' m = takeWhile (if n' >= n then (<= m) else (>= m))
-                                         (numericEnumFromThen n n')
+numericEnumFromThenTo n n' m = takeWhile p (numericEnumFromThen n n')
+                               where p | n' >= n   = (<= m)
+				       | otherwise = (>= m)
 
 primitive primShowsInt :: Int -> Int -> ShowS
 
@@ -848,12 +845,12 @@ realFloatToRational x = (m%1)*(b%1)^^n
 			      b     = floatRadix x
 
 primitive primDivFloat      :: Float -> Float -> Float
-primitive primDoubleToFloat :: Double -> Float
+primitive doubleToFloat     :: Double -> Float
 
 instance Fractional Float where
     (/)          = primDivFloat
     fromRational = primRationalToFloat
-    fromDouble   = primDoubleToFloat
+    fromDouble   = doubleToFloat
 
 primitive primDivDouble :: Double -> Double -> Double
 
@@ -884,13 +881,11 @@ rationalToRealFloat x = x'
 			     / fromInteger (denominator x))
        b     = floatRadix x'
 
-primitive primPiFloat :: Float
 primitive primSinFloat,  primAsinFloat, primCosFloat,
 	  primAcosFloat, primTanFloat,  primAtanFloat,
 	  primLogFloat,  primExpFloat,  primSqrtFloat :: Float -> Float
 
 instance Floating Float where
-    pi    = primPiFloat
     exp   = primExpFloat
     log   = primLogFloat
     sqrt  = primSqrtFloat
@@ -901,13 +896,11 @@ instance Floating Float where
     acos  = primAcosFloat
     atan  = primAtanFloat
 
-primitive primPiDouble :: Double
 primitive primSinDouble,  primAsinDouble, primCosDouble,
 	  primAcosDouble, primTanDouble,  primAtanDouble,
 	  primLogDouble,  primExpDouble,  primSqrtDouble :: Double -> Double
 
 instance Floating Double where
-    pi    = primPiDouble
     exp   = primExpDouble
     log   = primLogDouble
     sqrt  = primSqrtDouble
@@ -970,20 +963,20 @@ instance RealFloat Double where
     isIEEE      _ = False
 
 instance Enum Float where
-    toEnum         = primIntToFloat
-    fromEnum       = truncate
-    enumFrom       = numericEnumFrom
-    enumFromTo     = numericEnumFromTo
-    enumFromThen   = numericEnumFromThen
-    enumFromThenTo = numericEnumFromThenTo
+    toEnum		  = primIntToFloat
+    fromEnum		  = truncate
+    enumFrom		  = numericEnumFrom
+    enumFromThen	  = numericEnumFromThen
+    enumFromTo n m	  = numericEnumFromTo n (m+1/2)
+    enumFromThenTo n n' m = numericEnumFromThenTo n n' (m + (n'-n)/2)
 
 instance Enum Double where
-    toEnum         = primIntToDouble
-    fromEnum       = truncate
-    enumFrom       = numericEnumFrom
-    enumFromTo     = numericEnumFromTo
-    enumFromThen   = numericEnumFromThen
-    enumFromThenTo = numericEnumFromThenTo
+    toEnum		  = primIntToDouble
+    fromEnum		  = truncate
+    enumFrom		  = numericEnumFrom
+    enumFromThen	  = numericEnumFromThen
+    enumFromTo n m	  = numericEnumFromTo n (m+1/2)
+    enumFromThenTo n n' m = numericEnumFromThenTo n n' (m + (n'-n)/2)
 
 primitive primShowsFloat :: Int -> Float -> ShowS
 
@@ -1026,9 +1019,6 @@ const k _       = k
 (.)            :: (b -> c) -> (a -> b) -> (a -> c)
 (f . g) x       = f (g x)
 
---(.)          :: Functor f => (a -> b) -> (f a -> f b)
---(.)           = map
-
 flip           :: (a -> b -> c) -> b -> a -> c
 flip f x y      = f y x
 
@@ -1055,7 +1045,7 @@ type Rational              = Ratio Integer
 x % y                      = reduce (x * signum y) (abs y)
 
 reduce                    :: Integral a => a -> a -> Ratio a
-reduce x y | y == 0        = error "PreludeRatio.%: zero denominator"
+reduce x y | y == 0        = error "Ratio.%: zero denominator"
 	   | otherwise     = (x `quot` d) :% (y `quot` d)
 			     where d = gcd x y
 
@@ -1153,22 +1143,35 @@ null             :: [a] -> Bool
 null []           = True
 null (_:_)        = False
 
+(++)             :: [a] -> [a] -> [a]
+[]     ++ ys      = ys
+(x:xs) ++ ys      = x : (xs ++ ys)
+
+map              :: (a -> b) -> [a] -> [b]
+map f xs          = [ f x | x <- xs ]
+
+filter           :: (a -> Bool) -> [a] -> [a]
+filter p xs       = [ x | x <- xs, p x ]
+
+concat           :: [[a]] -> [a]
+concat            = foldr (++) []
+
 length           :: [a] -> Int
 length            = foldl' (\n _ -> n + 1) 0
 
 (!!)             :: [b] -> Int -> b
 (x:_)  !! 0       = x
 (_:xs) !! n | n>0 = xs !! (n-1)
-(_:_)  !! _       = error "PreludeList.!!: negative index"
-[]     !! _       = error "PreludeList.!!: index too large"
+(_:_)  !! _       = error "Prelude.!!: negative index"
+[]     !! _       = error "Prelude.!!: index too large"
 
 foldl            :: (a -> b -> a) -> a -> [b] -> a
 foldl f z []      = z
 foldl f z (x:xs)  = foldl f (f z x) xs
 
-foldl'           :: Eval a => (a -> b -> a) -> a -> [b] -> a
+foldl'           :: (a -> b -> a) -> a -> [b] -> a
 foldl' f a []     = a
-foldl' f a (x:xs) = strict (foldl' f) (f a x) xs
+foldl' f a (x:xs) = (foldl' f $! f a x) xs
 
 foldl1           :: (a -> a -> a) -> [a] -> a
 foldl1 f (x:xs)   = foldl f x xs
@@ -1209,25 +1212,26 @@ replicate        :: Int -> a -> [a]
 replicate n x     = take n (repeat x)
 
 cycle            :: [a] -> [a]
+cycle []          = error "Prelude.cycle: empty list"
 cycle xs          = xs' where xs'=xs++xs'
 
 take                :: Int -> [a] -> [a]
 take 0 _             = []
 take _ []            = []
 take n (x:xs) | n>0  = x : take (n-1) xs
-take _ _             = error "PreludeList.take: negative argument"
+take _ _             = error "Prelude.take: negative argument"
 
 drop                :: Int -> [a] -> [a]
 drop 0 xs            = xs
 drop _ []            = []
 drop n (_:xs) | n>0  = drop (n-1) xs
-drop _ _             = error "PreludeList.drop: negative argument"
+drop _ _             = error "Prelude.drop: negative argument"
 
 splitAt               :: Int -> [a] -> ([a], [a])
 splitAt 0 xs           = ([],xs)
 splitAt _ []           = ([],[])
 splitAt n (x:xs) | n>0 = (x:xs',xs'') where (xs',xs'') = splitAt (n-1) xs
-splitAt _ _            = error "PreludeList.splitAt: negative argument"
+splitAt _ _            = error "Prelude.splitAt: negative argument"
 
 takeWhile           :: (a -> Bool) -> [a] -> [a]
 takeWhile p []       = []
@@ -1244,8 +1248,9 @@ dropWhile p xs@(x:xs')
 span, break         :: (a -> Bool) -> [a] -> ([a],[a])
 span p []            = ([],[])
 span p xs@(x:xs')
-	 | p x       = let (ys,zs) = span p xs' in (x:ys,zs)
+	 | p x       = (x:ys, zs)
 	 | otherwise = ([],xs)
+                       where (ys,zs) = span p xs'
 break p              = span (not . p)
 
 lines     :: String -> [String]
@@ -1267,16 +1272,16 @@ unwords   :: [String] -> String
 unwords [] = []
 unwords ws = foldr1 (\w s -> w ++ ' ':s) ws
 
-reverse          :: [a] -> [a]
-reverse           = foldl (flip (:)) []
+reverse   :: [a] -> [a]
+reverse    = foldl (flip (:)) []
 
-and, or        :: [Bool] -> Bool
-and             = foldr (&&) True
-or              = foldr (||) False
+and, or   :: [Bool] -> Bool
+and        = foldr (&&) True
+or         = foldr (||) False
 
-any, all       :: (a -> Bool) -> [a] -> Bool
-any p           = or  . map p
-all p           = and . map p
+any, all  :: (a -> Bool) -> [a] -> Bool
+any p      = or  . map p
+all p      = and . map p
 
 elem, notElem    :: Eq a => a -> [a] -> Bool
 elem              = any . (==)
@@ -1332,11 +1337,8 @@ shows         = showsPrec 0
 read         :: Read a => String -> a
 read s        =  case [x | (x,t) <- reads s, ("","") <- lex t] of
 		      [x] -> x
-		      []  -> error "PreludeText.read: no parse"
-		      _   -> error "PreludeText.read: ambiguous parse"
-
-show         :: Show a => a -> String
-show x        = shows x ""
+		      []  -> error "Prelude.read: no parse"
+		      _   -> error "Prelude.read: ambiguous parse"
 
 showChar     :: Char -> ShowS
 showChar      = (:)
@@ -1360,9 +1362,9 @@ readParen b g = if b then mandatory else optional
 readField    :: Read a => String -> ReadS a
 readField m s0 = [ r | (t,  s1) <- lex s0, t == m,
                        ("=",s2) <- lex s1,
-                       r        <- readsPrec 10 s2 ]
+                       r        <- reads s2 ]
 
-lex                     :: ReadS String
+lex                    :: ReadS String
 lex ""                  = [("","")]
 lex (c:s) | isSpace c   = lex (dropWhile isSpace s)
 lex ('\'':s)            = [('\'':ch++"'", t) | (ch,'\'':t)  <- lexLitChar s,
@@ -1376,7 +1378,7 @@ lex ('"':s)             = [('"':str, t)      | (str,t) <- lexString s]
 
 			  lexStrItem ('\\':'&':s) = [("\\&",s)]
 			  lexStrItem ('\\':c:s) | isSpace c
-			      = [("\\&",t) | '\\':t <- [dropWhile isSpace s]]
+			      = [("",t) | '\\':t <- [dropWhile isSpace s]]
 			  lexStrItem s            = lexLitChar s
 
 lex (c:s) | isSingle c  = [([c],s)]
@@ -1388,7 +1390,7 @@ lex (c:s) | isSingle c  = [([c],s)]
 		where
 		isSingle c  =  c `elem` ",;()[]{}_`"
 		isSym c     =  c `elem` "!@#$%&*+./<=>?\\^|:-~"
-		isIdChar c  =  isAlphanum c || c `elem` "_'"
+		isIdChar c  =  isAlphaNum c || c `elem` "_'"
 
 		lexFracExp ('.':s) = [('.':ds++e,u) | (ds,t) <- lexDigits s,
 						      (e,u)  <- lexExp t    ]
@@ -1553,7 +1555,7 @@ instance Show (IO a) where
 primitive primbindIO   "rbindIO" :: IO a -> (a -> IO b) -> IO b
 primitive primretIO    "runitIO" :: a -> IO a
 primitive catch        "lbindIO" :: IO a -> (IOError -> IO a) -> IO a
-primitive fail         "lunitIO" :: IOError -> IO a
+primitive ioError      "lunitIO" :: IOError -> IO a
 primitive putChar		 :: Char -> IO ()
 primitive putStr		 :: String -> IO ()
 primitive getChar   		 :: IO Char
@@ -1576,9 +1578,9 @@ getLine    = do c <- getChar
 readIO          :: Read a => String -> IO a
 readIO s         = case [x | (x,t) <- reads s, ("","") <- lex t] of
                         [x] -> return x
-                        []  -> fail (userError "PreludeIO.readIO: no parse")
-                        _   -> fail (userError 
-                                      "PreludeIO.readIO: ambiguous parse")
+                        []  -> ioError (userError "PreludeIO.readIO: no parse")
+                        _   -> ioError (userError 
+                                       "PreludeIO.readIO: ambiguous parse")
 
 readLn          :: Read a => IO a
 readLn           = do l <- getLine
@@ -1594,40 +1596,81 @@ interact  :: (String -> String) -> IO ()
 interact f = getContents >>= (putStr . f)
 
 instance Functor IO where
-    map f x = x >>= (return . f)
+    fmap f x = x >>= (return . f)
 
 instance Monad IO where
     (>>=)  = primbindIO
     return = primretIO
 
+
 -- Hooks for primitives: -----------------------------------------------------
 -- Do not mess with these!
 
-newtype IO a = IO ((IOError -> IOResult a) -> (a -> IOResult a) -> IOResult a)
-data IOResult a 
-  = Hugs_ExitWith Int
-  | Hugs_SuspendThread
-  | Hugs_Error    IOError
-  | Hugs_Return   a
+data Addr     -- builtin datatype of C pointers
+data Word     -- builtin datatype of unsigned ints
+data ForeignObj
+data StablePtr a
+
+primitive unsafeCoerce "primUnsafeCoerce" :: a -> b
+
+data Obj = Obj
+
+toObj :: a -> Obj
+toObj   = unsafeCoerce
+
+fromObj :: Obj -> a
+fromObj = unsafeCoerce
+
+
+newtype IO a = IO ((IOError -> IOResult) -> (a -> IOResult) -> IOResult)
+data IOResult 
+  = Hugs_ExitWith    Int
+  | Hugs_Error       IOError
+  | Hugs_ForkThread  IOResult IOResult
+  | Hugs_DeadThread
+  | Hugs_YieldThread IOResult
+  | Hugs_Return      Obj
+
+data IOFinished a
+  = Finished_ExitWith Int
+  | Finished_Error    IOError
+  | Finished_Return   a
 
 hugsPutStr :: String -> IO ()
 hugsPutStr  = putStr
 
 hugsIORun  :: IO a -> Either Int a
-hugsIORun m = performIO (runAndShowError m)
+hugsIORun m = 
+  case basicIORun (runAndShowError m) of
+    Finished_ExitWith i -> Left i
+    Finished_Error    _ -> Left 1
+    Finished_Return   a -> Right a
  where
-  performIO       :: IO a -> Either Int a
-  performIO (IO m) = case m Hugs_Error Hugs_Return of
-	             Hugs_Return a   -> Right a
-		     Hugs_ExitWith e -> Left  e
-		     _               -> Left  1
-
   runAndShowError :: IO a -> IO a
   runAndShowError m =
     m `catch` \err -> do 
 	putChar '\n'
 	putStr (ioeGetErrorString err)
-	primExitWith 1 -- alternatively: (IO (\f s -> Hugs_SuspendThread))
+	primExitWith 1
+
+basicIORun :: IO a -> IOFinished a
+basicIORun (IO m) = loop [m Hugs_Error (Hugs_Return . toObj)]
+
+threadToIOResult :: IO a -> IOResult
+threadToIOResult (IO m) = m Hugs_Error (const Hugs_DeadThread)
+
+-- This is the queue of *runnable* threads.
+-- There may be blocked processes inside
+loop :: [IOResult] -> IOFinished a
+loop []                      = error "no more threads (deadlock?)"
+loop [Hugs_Return   a]       = Finished_Return (fromObj a)
+loop (Hugs_Return   a:r)     = loop (r ++ [Hugs_Return a])
+loop (Hugs_Error    e:_)     = Finished_Error  e
+loop (Hugs_ExitWith i:_)     = Finished_ExitWith i
+loop (Hugs_DeadThread:r)     = loop r
+loop (Hugs_ForkThread a b:r) = loop (a:b:r)
+loop (Hugs_YieldThread a:r)  = loop (r ++ [a])
+loop _                       = undefined
 
 primExitWith     :: Int -> IO a
 primExitWith c    = IO (\ f s -> Hugs_ExitWith c)
