@@ -3,7 +3,7 @@
 -- normally be more selective than this when it comes to specifying the
 -- modules that a particular program depends on.
 --
--- Suitable for use with Hugs 1.4.
+-- Suitable for use with Hugs 98
 -----------------------------------------------------------------------------
 
 module HugsLibs where
@@ -17,7 +17,6 @@ import AnsiScreen
 import AnsiInteract
 import IOExtensions
 import Sequence
-import Random
 import ListUtils
 import Dynamic
 
