@@ -12,16 +12,19 @@
  * specifications will be specified for these operations in the final
  * version of the 1.3 I/O specification, so be prepared for changes.
  *
- * Copyright (c) The University of Nottingham and Yale University, 1994-1997.
- * All rights reserved. See NOTICE for details and conditions of use etc...
- * Hugs version 1.4, December 1997
+ * The Hugs 98 system is Copyright (c) Mark P Jones, Alastair Reid, the
+ * Yale Haskell Group, and the Oregon Graduate Institute of Science and
+ * Technology, 1994-1999, All rights reserved.  It is distributed as
+ * free software under the license in the file "License", which is
+ * included in the distribution.
  *
  * $RCSfile: iomonad.c,v $
- * $Revision: 1.27 $
- * $Date: 1997/12/12 15:09:01 $
+ * $Revision: 1.11 $
+ * $Date: 2001/02/14 12:15:05 $
  * ------------------------------------------------------------------------*/
  
 Name nameIORun;			        /* run IO code                     */
+Name nameIOBind;		        /* bind IO code                    */
 Name namePutStr;		        /* Prelude.putStr                  */
 
 static Name namePass;			/* auxiliary:: \f b c a -> f a b c */
@@ -41,8 +44,8 @@ static Void iomonadControl(what)
 Int what; {
     switch (what) {
 	case INSTALL : 
-	               setCurrModule(modulePrelude);
-#define pFun(n,s,t)    addPrim(0,n=newName(findText(s),NIL),t,NIL)
+		       setCurrModule(modulePrelude);
+#define pFun(n,s,t)    addPrim(0,n=newName(findText(s),NIL),t,modulePrelude,NIL)
 		       pFun(namePass,	 "_pass",    "passIO");
 #if    IO_HANDLES
 		       pFun(nameHreader, "_hreader", "hreader");
@@ -50,13 +53,14 @@ Int what; {
 #undef pFun
 #define predef(nm,str) nm=newName(findText(str),NIL); name(nm).defn=PREDEFINED
 		       predef(nameIORun,    "hugsIORun");
+		       predef(nameIOBind,   "primbindIO");
 		       predef(namePutStr,   "hugsPutStr");
 #undef predef
-	               break;
+		       break;
 
-        case RESET   : 
+	case RESET   : 
 #if IO_HANDLES
-                       if (writingFile) {
+		       if (writingFile) {
 			   fclose(writingFile);
 			   writingFile = 0;
 		       }
@@ -83,6 +87,9 @@ PROTO_PRIM(primShowIOError);
 PROTO_PRIM(primUserError);
 PROTO_PRIM(primIsUserErr);
 
+PROTO_PRIM(primIsSearchErr);
+PROTO_PRIM(primIsNameErr);
+
 PROTO_PRIM(primGetCh);
 PROTO_PRIM(primGetChar);
 PROTO_PRIM(primPutChar);
@@ -120,6 +127,8 @@ PROTO_PRIM(primReadBinaryFile);
 PROTO_PRIM(primWriteBinaryFile);
 PROTO_PRIM(primAppendBinaryFile);
 PROTO_PRIM(primIsIllegal);
+PROTO_PRIM(primIsWriteErr);
+PROTO_PRIM(primIsEOFError);
 PROTO_PRIM(primIsUnsupported);
 PROTO_PRIM(primGetHandle);
 PROTO_PRIM(primGetFileName);
@@ -132,6 +141,37 @@ PROTO_PRIM(primAssignRef);
 PROTO_PRIM(primEqRef);
 #endif
 
+PROTO_PRIM(primMakeSP);
+PROTO_PRIM(primDerefSP);
+PROTO_PRIM(primFreeSP);
+
+PROTO_PRIM(primMakeFO);
+PROTO_PRIM(primWriteFO);
+PROTO_PRIM(primEqFO);
+
+#if GC_WEAKPTRS
+PROTO_PRIM(primMakeWeakPtr);
+PROTO_PRIM(primDerefWeakPtr);
+PROTO_PRIM(primWeakPtrEq);
+PROTO_PRIM(primMkWeak);
+PROTO_PRIM(primDeRefWeak);
+PROTO_PRIM(primReplaceFinalizer);
+PROTO_PRIM(primFinalize);
+PROTO_PRIM(primRunFinalizer);
+PROTO_PRIM(primFinalizerWaiting);
+#endif
+
+#if STABLE_NAMES
+PROTO_PRIM(primMakeSN);
+PROTO_PRIM(primDerefSN);
+PROTO_PRIM(primHashSN);
+PROTO_PRIM(primEqSN);
+#endif
+
+#ifdef HSCRIPT
+PROTO_PRIM(primGetCurrentScript);
+#endif
+
 static struct primitive iomonadPrimTable[] = {
   {"lunitIO",		3, primLunit},
   {"runitIO",		3, primRunit},
@@ -155,6 +195,10 @@ static struct primitive iomonadPrimTable[] = {
   {"primShowIOError",	1, primShowIOError},
   {"userError",		1, primUserError},
   {"isUserError",	1, primIsUserErr},
+  /* non-standard tests */
+  {"hugsIsSearchErr",	1, primIsSearchErr},
+  {"hugsIsNameErr",	        1, primIsNameErr},
+  /* end non-standard tests */
 
 #if IO_HANDLES
   {"hGetChar",		3, primHGetChar},
@@ -185,12 +229,15 @@ static struct primitive iomonadPrimTable[] = {
   {"writeBinaryFile",	4, primWriteBinaryFile},
   {"appendBinaryFile",	4, primAppendBinaryFile},
   {"isAlreadyExistsError", 1, primIsUnsupported},
-  {"isDoesNotExistError",  1, primIsUnsupported},
+  {"isDoesNotExistError",  1, primIsSearchErr},
   {"isAlreadyInUseError",  1, primIsUnsupported},
   {"isFullError",	   1, primIsUnsupported},
-  {"isEOFError",	   1, primIsUnsupported},
+  {"isEOFError",	   1, primIsEOFError},
   {"isIllegalOperation",   1, primIsIllegal},
   {"isPermissionError",	   1, primIsUnsupported},
+  /* non-standard tests */
+  {"hugsIsWriteErr",       1, primIsWriteErr},
+  /* end non-standard tests */
   {"ioeGetHandle",	1, primGetHandle},
   {"ioeGetFileName",	1, primGetFileName},
 #endif
@@ -202,6 +249,37 @@ static struct primitive iomonadPrimTable[] = {
   {"eqRef",		2, primEqRef},
 #endif
 
+  {"makeStablePtr",	3, primMakeSP},
+  {"deRefStablePtr",	3, primDerefSP},
+  {"freeStablePtr",	3, primFreeSP},
+
+  {"makeForeignObj",	4, primMakeFO},
+  {"writeForeignObj",	4, primWriteFO},
+  {"eqForeignObj",	2, primEqFO},
+
+#if GC_WEAKPTRS
+  {"makeWeakPtr",       3, primMakeWeakPtr},
+  {"derefWeakPtr",      3, primDerefWeakPtr},
+  {"weakPtrEq",		2, primWeakPtrEq},
+  {"mkWeak",		5, primMkWeak},
+  {"deRefWeak",		3, primDeRefWeak},
+  {"replaceFinalizer",	4, primReplaceFinalizer},
+  {"finalize",		3, primFinalize},
+  {"runFinalizer",	2, primRunFinalizer},
+  {"finalizerWaiting",	2, primFinalizerWaiting},
+#endif
+
+#if STABLE_NAMES
+  {"makeStableName",	3, primMakeSN},
+  {"deRefStableName",	1, primDerefSN},
+  {"hashStableName",	1, primHashSN},
+  {"eqStableName",	2, primEqSN},
+#endif
+
+#ifdef HSCRIPT
+  {"getCurrentScript",  2, primGetCurrentScript},
+#endif
+
   {0,			0, 0}
 };
 
@@ -302,6 +380,20 @@ primFun(primIsUserErr) {		/* :: IOError -> Bool        	   */
     BoolResult(whnfHead==nameUserErr);
 }
 
+primFun(primIsNameErr) {		/* :: IOError -> Bool		   */
+    eval(primArg(1));
+    BoolResult(whnfHead==nameNameErr);
+}
+
+primFun(primIsSearchErr) {		/* :: IOError -> Bool		   */
+    eval(primArg(1));
+    BoolResult(whnfHead==nameSearchErr);
+}
+
+primFun(primIsUnsupported) {		/* :: IOError -> Bool		   */
+    updateRoot(nameFalse);
+}
+
 /* --------------------------------------------------------------------------
  * Misc.
  * ------------------------------------------------------------------------*/
@@ -326,17 +418,17 @@ primFun(primGetRandomSeed) {		/* generate a random seed          */
 }
 
 #endif
-                                     
+				     
 primFun(primGetEnv) {                 /* primGetEnv :: String -> IO String */
     String s = evalName(IOArg(1));    /* Eval name	                   */
     String r;
     if (!s) {			      /* check for valid name		   */
 	IOFail(ap(nameNameErr,IOArg(1)));
     } else if ((r = getenv(s))!=0) {    
-        pushString(r);
+	pushString(r);
 	IOReturn(pop());
     } else {
-        IOFail(ap(nameSearchErr,IOArg(1)));
+	IOFail(ap(nameSearchErr,IOArg(1)));
     }
 }
 
@@ -353,13 +445,17 @@ primFun(primGetEnv) {                 /* primGetEnv :: String -> IO String */
 # define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
 #endif
 
-                                         
+					 
 primFun(primSystem) {                   /* primSystem :: String -> IO Int  */
     String s = evalName(IOArg(1));	/* Eval name	                   */
     Int r;
     if (s) {				/* check for valid string          */
-        r = system(s);
-        IOReturn(mkInt(WEXITSTATUS(r)));
+#if HAVE_MACSYSTEM
+        r = macsystem(s);
+#else
+	r = system(s);
+#endif
+	IOReturn(mkInt(WEXITSTATUS(r)));
     } else {
 	IOFail(ap(nameNameErr,IOArg(1)));
     }
@@ -387,10 +483,10 @@ primFun(primArgv) {                     /* primArgv :: Int -> IO String    */
     Int i;
     IntArg(i,3);
     if (0 <= i && i < hugsArgc) {
-        pushString(hugsArgv[i]);
-        IOReturn(pop());
+	pushString(hugsArgv[i]);
+	IOReturn(pop());
     } else {
-        IOFail(nameIllegal);
+	IOFail(nameIllegal);
     }
 }
     
@@ -402,12 +498,18 @@ primFun(primGetCh) {			/* Get character from stdin wo/echo*/
     IOReturn(mkChar(readTerminalChar()));
 }
 
+#if __MWERKS__ && macintosh
+primFun(primGetChar) {			 /* Metrowerks console has no NO_ECHO mode. */
+    IOReturn(mkChar(readTerminalChar()));
+}
+#else
 primFun(primGetChar) {			/* Get character from stdin w/ echo*/
     Char c = readTerminalChar();
     putchar(c);
     fflush(stdout);
     IOReturn(mkChar(c));
 }
+#endif
 
 primFun(primPutChar) {			/* print character on stdout	   */
     eval(pop());
@@ -421,7 +523,7 @@ primFun(primPutStr) {			/* print string on stdout	   */
     eval(pop());			/* included here for speed	   */
     while (whnfHead==nameCons) {
 	eval(top());
-	checkChar(top());
+	checkChar();
 	putchar(charOf(whnfHead));
 #if FLUSHEVERY
 	fflush(stdout);
@@ -449,7 +551,7 @@ primFun(primPutStr) {			/* print string on stdout	   */
     IOReturn((e)?nameTrue:nameFalse)
 
 primFun(primHGetChar) {			/* Read character from handle	   */
-    Int h = 0;
+    Int h;
     HandleArg(h,3);
     if (handles[h].hmode&HREAD) {
 	Char c = (h==HSTDIN ? readTerminalChar() : getc(handles[h].hfp));
@@ -461,8 +563,8 @@ primFun(primHGetChar) {			/* Read character from handle	   */
 }
 
 primFun(primHPutChar) {			/* print character on handle	   */
-    Int  h = 0;
     Char c = 0;
+    Int  h;
     HandleArg(h,4);
     CharArg(c,3);
     if (handles[h].hmode&(HWRITE|HAPPEND)) {
@@ -473,9 +575,10 @@ primFun(primHPutChar) {			/* print character on handle	   */
 }
 
 primFun(primHPutStr) {			/* print string on handle	   */
-    Int  h = 0;
+    Int h;
     HandleArg(h,4);
-    drop();
+    push(primArg(3));
+    primArg(3) = NIL;
     if (handles[h].hmode&(HWRITE|HAPPEND)) {
 	blackHoleRoot();
 	eval(pop());
@@ -496,7 +599,7 @@ primFun(primHPutStr) {			/* print string on handle	   */
 }
 
 primFun(primHreader) {			/* read String from a handle 	   */
-    Int h=0;                            /* Handle -> String                */
+    Int h;                              /* Handle -> String                */
     HandleArg(h,1);
     if (handles[h].hmode&HSEMICLOSED) {	/* read requires semi-closed handle*/
 	Int c = (h==HSTDIN ? readTerminalChar() : getc(handles[h].hfp));
@@ -510,7 +613,7 @@ primFun(primHreader) {			/* read String from a handle 	   */
 }
 
 primFun(primHContents) {		/* hGetContents :: Handle -> IO Str*/
-    Int h = 0;
+    Int h;
     HandleArg(h,3);
     if ((handles[h].hmode&HREAD)==0) {	/* must have readable handle	   */
 	IOFail(nameIllegal);
@@ -581,7 +684,7 @@ primFun(primStderr) {			/* Standard error handle	   */
 
 /* NOTE: this doesn't implement the Haskell 1.3 semantics */
 primFun(primHIsEOF) {			/* Test for end of file on handle  */
-    Int h = 0;
+    Int h;
     HandleArg(h,3);
     if (handles[h].hmode!=HCLOSED) {
 	IOReturn(feof(handles[h].hfp) ? nameTrue : nameFalse);
@@ -591,7 +694,7 @@ primFun(primHIsEOF) {			/* Test for end of file on handle  */
 }
 
 primFun(primHFlush) {			/* Flush handle			   */
-    Int h = 0;
+    Int h;
     HandleArg(h,3);
     if (handles[h].hmode!=HCLOSED) {
 	fflush(handles[h].hfp);
@@ -602,7 +705,7 @@ primFun(primHFlush) {			/* Flush handle			   */
 }
 
 primFun(primHClose) {			/* Close handle                   */
-    Int h = 0;
+    Int h;
     HandleArg(h,3);
     if (handles[h].hmode!=HCLOSED) {
 	if (h>HSTDERR && handles[h].hfp)
@@ -615,72 +718,59 @@ primFun(primHClose) {			/* Close handle                   */
 }
 
 primFun(primHGetPosn) {			/* Get file position               */
-    Int h = 0;
+    Int h;
     HandleArg(h,3);
     if (handles[h].hmode!=HCLOSED) {
-#if HAVE_FGETPOS
-	fpos_t pos;
-	if (fgetpos(handles[h].hfp,&pos)) {
-	    IOReturn(mkInt((Int)pos));
-	}
-#elif HAVE_FTELL
-	/* A reasonable approximation for pre-ANSI compilers */
+#if HAVE_FTELL
 	long pos = ftell(handles[h].hfp);
 	IOReturn(mkInt((Int)pos));
 #else
-        /* deliberate fall through to IOFail */
+	/* deliberate fall through to IOFail */
 #endif
     }
     IOFail(nameIllegal);
 }
 
 primFun(primHSetPosn) {			/* Set file position               */
-    Int    h   = 0;
-#if HAVE_FSETPOS
-    fpos_t pos = 0;
-#elif HAVE_FSEEK
+#if HAVE_FSEEK
     long   pos = 0;
 #endif
+    Int    h;
     HandleArg(h,3);
     IntArg(pos,4);
     if (handles[h].hmode!=HCLOSED) {
-#if HAVE_FSETPOS
-	if (fsetpos(handles[h].hfp,&pos)) {
-	    IOReturn(nameUnit);
-	}
-#elif HAVE_FSEEK
-	/* A reasonable approximation for pre-ANSI compilers */
+#if HAVE_FSEEK
 	if (fseek(handles[h].hfp,pos,SEEK_SET)) {
 	    IOReturn(nameUnit);
 	}
 #else
-        /* deliberate fall through to IOFail */
+	/* deliberate fall through to IOFail */
 #endif
     }
     IOFail(nameIllegal);
 }
 
 primFun(primHIsOpen) {			/* Test is handle open             */
-    Int    h   = 0;
+    Int h;
     HandleArg(h,3);
     IOBoolResult(handles[h].hmode!=HCLOSED 
 		 && handles[h].hmode!=HSEMICLOSED);
 }
 
 primFun(primHIsClosed) {		/* Test is handle closed           */
-    Int    h   = 0;
+    Int h;
     HandleArg(h,3);
     IOBoolResult(handles[h].hmode==HCLOSED);
 }
 
 primFun(primHIsReadable) {		/* Test is handle readable         */
-    Int    h   = 0;
+    Int h;
     HandleArg(h,3);
     IOBoolResult(handles[h].hmode&HREAD);
 }
 
 primFun(primHIsWritable) {		/* Test is handle writable         */
-    Int    h   = 0;
+    Int h;
     HandleArg(h,3);
     IOBoolResult(handles[h].hmode&HWRITE);
 }
@@ -693,30 +783,33 @@ primFun(primEqHandle) {			/* Test for handle equality        */
 }
 
 primFun(primReadFile) {			/* read file as lazy string	   */
-    String s   = evalName(IOArg(1));	/* Eval and check filename	   */
-    Cell   hnd = NIL;
+    String s = evalName(IOArg(1));	/* Eval and check filename	   */
     if (!s) {
 	IOFail(ap(nameNameErr,IOArg(1)));
-    } else if (isNull(hnd=openHandle(s,HREAD,FALSE))) {
-	IOFail(ap(nameSearchErr,IOArg(1)));
-    } else {
+    }
+    else {
+	Cell hnd = openHandle(s,HREAD,FALSE);
+	if (isNull(hnd)) {
+	    IOFail(ap(nameSearchErr,IOArg(1)));
+	}
 	handles[intValOf(hnd)].hmode = HSEMICLOSED;
 	IOReturn(ap(nameHreader,hnd));
     }
 }
 
 primFun(primReadBinaryFile) {		/* read file as lazy string	   */
-    String s   = evalName(IOArg(1));	/* Eval and check filename	   */
-    Cell   hnd = NIL;
+    String s = evalName(IOArg(1));	/* Eval and check filename	   */
     if (!s) {
 	IOFail(ap(nameNameErr,IOArg(1)));
     }
-    hnd=openHandle(s,HREAD,TRUE);
-    if (isNull(hnd)) {
-	IOFail(ap(nameSearchErr,IOArg(1)));
+    else {
+	Cell hnd = openHandle(s,HREAD,TRUE);
+	if (isNull(hnd)) {
+	    IOFail(ap(nameSearchErr,IOArg(1)));
+	}
+	handles[intValOf(hnd)].hmode = HSEMICLOSED;
+	IOReturn(ap(nameHreader,hnd));
     }
-    handles[intValOf(hnd)].hmode = HSEMICLOSED;
-    IOReturn(ap(nameHreader,hnd));
 }
 
 primFun(primWriteFile) {		/* write string to specified file  */
@@ -735,26 +828,33 @@ primFun(primAppendBinaryFile) {		/* append string to specified file */
     fwritePrim(root,TRUE,TRUE);
 }
 
+#if HAVE_GETFINFO
+extern int access(char *fileName, int dummy);
+#endif
+
 static Void local fwritePrim(root,append,binary)/* Auxiliary function for  */
 StackPtr root;					/* writing/appending to	   */
 Bool     append; 				/* an output file	   */
 Bool     binary; {
     String mode = binary ? (append ? "ab" : "wb")
-                         : (append ? "a"  : "w");
+			 : (append ? "a"  : "w");
     String s    = evalName(IOArg(2));		/* Eval and check filename */
     if (!s) {
 	IOFail(ap(nameNameErr,IOArg(2)));
-    } else if (append && access(s,0)!=0) {	/* Check that file exists  */
+    }
+    else if (append && access(s,0)!=0) {	/* Check that file exists  */
 	IOFail(ap(nameSearchErr,IOArg(2)));
-    } else if ((writingFile=fopen(s,mode))==0) {/* Open file for writing   */
+    }
+    else if ((writingFile=fopen(s,mode))==0) {	/* Open file for writing   */
 	IOFail(ap(nameWriteErr,IOArg(2)));
-    } else {					/* Output characters	   */
+    }
+    else {					/* Output characters	   */
 	blackHoleRoot();
 	drop();
 	eval(pop());
 	while (whnfHead==nameCons) {
 	    eval(top());
-	    checkChar(top());
+	    checkChar();
 	    fputc(charOf(whnfHead),writingFile);
 	    drop();
 	    eval(pop());
@@ -770,8 +870,14 @@ primFun(primIsIllegal) {		/* :: IOError -> Bool		   */
     BoolResult(whnfHead==nameIllegal);
 }
 
-primFun(primIsUnsupported) {		/* :: IOError -> Bool		   */
-    updateRoot(nameFalse);
+primFun(primIsWriteErr) {		/* :: IOError -> Bool		   */
+    eval(primArg(1));
+    BoolResult(whnfHead==nameWriteErr);
+}
+
+primFun(primIsEOFError) {		/* :: IOError -> Bool		   */
+    eval(primArg(1));
+    BoolResult(whnfHead==nameEOFErr);
 }
 
 primFun(primGetHandle) {		/* :: IOError -> Maybe Handle	   */
@@ -788,6 +894,7 @@ primFun(primGetFileName) {		/* :: IOError -> Maybe FilePath	   */
     else
 	updateRoot(nameNothing);
 }
+
 #endif /* IO_HANDLES */
 
 /* --------------------------------------------------------------------------
@@ -829,5 +936,229 @@ primFun(primEqRef) {			/* Ref a -> Ref a -> Bool	   */
 }
 #endif
 
+/* --------------------------------------------------------------------------
+ * Stable Pointers
+ * ------------------------------------------------------------------------*/
+
+primFun(primMakeSP) {			/* a -> IO (StablePtr a)	   */
+    Int sp = mkStablePtr(IOArg(1));
+    if (sp > 0) {
+	IOReturn(mkInt(sp));
+    } else {
+	IOFail(nameIllegal); 
+    }
+}
+
+primFun(primDerefSP) {			/* StablePtr a -> IO a   	   */
+    eval(IOArg(1));
+    IOReturn(derefStablePtr(whnfInt));
+}
+
+primFun(primFreeSP) {			/* StablePtr a -> IO ()   	   */
+    eval(IOArg(1));
+    freeStablePtr(whnfInt);
+    IOReturn(nameUnit);
+}
+    
+/* --------------------------------------------------------------------------
+ * Foreign Objects
+ * ------------------------------------------------------------------------*/
+
+#if CHECK_TAGS
+#define checkForeign() if (MPCELL != whatIs(whnfHead)) internal("ForeignObj expected")
+#else
+#define checkForeign() /* do nothing */
+#endif
+
+primFun(primMakeFO) {			/* a -> IO (Ref a)		   */
+    Pointer addr = 0;
+    Void (*free)(Pointer) = 0;
+    eval(IOArg(2));
+    addr = ptrOf(whnfHead);
+    eval(IOArg(1));
+    free = (Void (*)(Pointer))ptrOf(whnfHead);
+    IOReturn(mkMallocPtr(addr,free));
+}
+
+primFun(primWriteFO) {		/* ForeignObj -> Addr -> IO ()		   */
+    Cell mp = NIL;
+    eval(IOArg(2));
+    checkForeign();
+    mp = whnfHead;
+    eval(IOArg(1));
+    derefMP(mp) = ptrOf(whnfHead);
+    IOReturn(nameUnit);
+}
+
+primFun(primEqFO) {			/* ForeignObj -> ForeignObj -> Bool*/
+    eval(primArg(2));
+    checkForeign();
+    push(whnfHead);
+    eval(primArg(1));
+    checkForeign();
+    updateRoot(pop()==whnfHead ? nameTrue : nameFalse);
+}
+
+#if STABLE_NAMES
+/* --------------------------------------------------------------------------
+ * Stable Names
+ * ------------------------------------------------------------------------*/
+
+primFun(primMakeSN) {		/* a -> IO (StableName a)		   */
+    IOReturn(ap(STABLENAME,IOArg(1)));
+}
+
+primFun(primDerefSN) {		/* StableName a -> a			   */
+    eval(primArg(1));
+    updateRoot(snd(whnfHead));
+}
+
+primFun(primHashSN) {		/* StableName a -> Int			   */
+    eval(primArg(1));
+    updateRoot(mkInt(whnfHead));
+}
+primFun(primEqSN) {		/* StableName a -> StableName a -> Bool	   */
+    eval(primArg(2));
+    push(whnfHead);
+    eval(primArg(1));
+    updateRoot(pop()==whnfHead ? nameTrue : nameFalse);
+}
+#endif
+
+#if GC_WEAKPTRS
+/* --------------------------------------------------------------------------
+ * Weak Pointers
+ * ------------------------------------------------------------------------*/
+
+#if CHECK_TAGS
+#define checkWeak() if(WEAKCELL!=whatIs(whnfHead)) internal("weakPtr expected");
+#else
+#define checkWeak() /* do nothing */
+#endif
+
+primFun(primMakeWeakPtr) {		/* a -> IO (Weak a)		   */
+    assert(isGenPair(IOArg(1)));	/* (Sadly, this may not be true)   */
+    IOReturn(mkWeakPtr(IOArg(1)));	/* OLD ... retire soon		   */
+}
+
+primFun(primDerefWeakPtr) {		/* Weak a -> IO (Maybe a)	   */
+    eval(IOArg(1));			/* OLD ... retire soon		   */
+    checkWeak();
+    if (isNull(derefWeakPtr(whnfHead))) {
+	IOReturn(nameNothing);
+    } else {
+	IOReturn(ap(nameJust,derefWeakPtr(whnfHead)));
+    }
+}
+
+primFun(primWeakPtrEq) {		/* Weak a -> Weak a -> Bool	   */
+    eval(primArg(2));
+    push(whnfHead);
+    eval(primArg(1));
+    updateRoot(pop()==whnfHead ? nameTrue : nameFalse);
+}
+
+primFun(primMkWeak) {			/* k -> v -> Maybe (IO ())	   */
+    Cell w = NIL;			/*		    -> IO (Weak v) */
+    eval(IOArg(1));
+    if (whnfHead==nameJust) {		/* Look for finalizer		   */
+	w = pop();
+    }
+    w		     = ap(NIL,ap(NIL,ap(NIL,w)));
+    fst(snd(w))      = IOArg(3);
+    fst(snd(snd(w))) = IOArg(2);
+    liveWeakPtrs     = cons(w,liveWeakPtrs);
+    fst(w)           = WEAKFIN;
+    IOReturn(w);
+}
+
+primFun(primDeRefWeak) {		/* Weak v -> IO (Maybe v)	   */
+    eval(IOArg(1));
+    if (whatIs(whnfHead)!=WEAKFIN) {
+	internal("primDeRefWeak");
+    }
+    if (nonNull(snd(whnfHead))) {
+	IOReturn(ap(nameJust,fst(snd(snd(whnfHead)))));
+    } else {
+	IOReturn(nameNothing);
+    }
+}
+
+primFun(primReplaceFinalizer) {		/* Weak v -> Maybe (IO ())	   */
+					/*	-> IO (Maybe (IO ()))	   */
+    eval(IOArg(1));			/* Grab new finalizer ...	   */
+    if (whnfHead!=nameJust) {
+	push(NIL);
+    }
+    eval(IOArg(2));			/* Get weak pointer ...		   */
+    if (whatIs(whnfHead)!=WEAKFIN) {
+	internal("primReplaceFinalizer");
+    } else if (nonNull(snd(whnfHead))) {/* ... and replace finalizer	   */
+	Cell oldfin = snd(snd(snd(whnfHead)));
+	snd(snd(snd(whnfHead))) = pop();
+	if (nonNull(oldfin)) {
+	    IOReturn(ap(nameJust,oldfin));
+	}
+    }
+    IOReturn(nameNothing);
+}
+
+primFun(primFinalize) {			/* Weak v -> IO ()		   */
+    eval(IOArg(1));			/* Bring weak pointer to an early  */
+    if (whatIs(whnfHead)!=WEAKFIN) {	/* end ...			   */
+	internal("primFinalize");
+    } else if (nonNull(snd(whnfHead))) {
+	Cell wp = whnfHead;
+	Cell vf = snd(snd(wp));
+	if (isPair(vf)) {
+	    if (nonNull(snd(vf))) {
+		fst(vf)    = snd(vf);
+		snd(vf)    = finalizers;
+		finalizers = vf;
+	    }
+	    fst(snd(wp)) = NIL;
+	    snd(snd(wp)) = NIL;
+	    snd(wp)      = NIL;
+	}
+	liveWeakPtrs = removeCell(wp,liveWeakPtrs);
+    }
+    IOReturn(nameUnit);
+}
+
+primFun(primRunFinalizer) {		/* IO ()			   */
+    if (isNull(finalizers)) {
+	IOReturn(nameUnit);
+    } else {
+	updapRoot(ap(hd(finalizers),primArg(2)),primArg(1));
+	finalizers = tl(finalizers);
+	return;
+    }
+}
+
+primFun(primFinalizerWaiting) {		/* IO Boolean			   */
+    if (isNull(finalizers)) {
+	IOReturn(nameFalse);
+    } else {
+	IOReturn(nameTrue);
+    }
+}
+#endif /* GC_WEAKPTRS */
+
+#if HSCRIPT
+#if EMBEDDED
+extern void* getCurrentScript(void);
+
+primFun(primGetCurrentScript) {  /* IO Int */
+    IOReturn( mkInt( (int)getCurrentScript() ) );
+}
+
+#else
+ 
+primFun(primGetCurrentScript) {  /* IO Int */
+    IOReturn( mkInt( 0 ) );
+}
+
+#endif /* EMBEDDED */
+#endif /* HSCRIPT */
 
 /*-------------------------------------------------------------------------*/
