@@ -2,13 +2,15 @@
  * Defines storage datatypes: Text, Name, Module, Tycon, Cell, List, Pair,
  * Triple, ...
  *
- * Copyright (c) The University of Nottingham and Yale University, 1994-1997.
- * All rights reserved. See NOTICE for details and conditions of use etc...
- * Hugs version 1.4, December 1997
+ * The Hugs 98 system is Copyright (c) Mark P Jones, Alastair Reid, the
+ * Yale Haskell Group, and the Oregon Graduate Institute of Science and
+ * Technology, 1994-1999, All rights reserved.  It is distributed as
+ * free software under the license in the file "License", which is
+ * included in the distribution.
  *
  * $RCSfile: storage.h,v $
- * $Revision: 1.40 $
- * $Date: 1997/12/12 15:09:07 $
+ * $Revision: 1.16 $
+ * $Date: 2001/02/14 12:15:05 $
  * ------------------------------------------------------------------------*/
 
 /* --------------------------------------------------------------------------
@@ -32,6 +34,7 @@ typedef Int	     Module;			 /* module 	           */
 typedef Cell	     Tycon;			 /* type constructor	   */
 typedef Cell	     Type;			 /* type expression	   */
 typedef Cell	     Kind;			 /* kind expression	   */
+typedef Cell	     Kinds;			 /* list of kinds	   */
 typedef Cell	     Constr;			 /* constructor expression */
 typedef Cell	     Name;			 /* named value 	   */
 typedef Void	     (*Prim) Args((StackPtr));	 /* primitive function	   */
@@ -46,6 +49,10 @@ typedef FloatImpType Float;			 /* implementation of Float*/
 #if TREX
 typedef Cell	     Ext;			 /* extension label	   */
 #endif
+#if OBSERVATIONS
+typedef Int          Observe;                    /* observation            */
+typedef Int          Breakpt;                    /* Breakpoint             */
+#endif
 
 /* --------------------------------------------------------------------------
  * Text storage:
@@ -58,12 +65,9 @@ extern	Text	     findText		Args((String));
 extern	Text         inventText		Args((Void));
 extern  Text	     inventDictText	Args((Void));
 extern  Bool	     inventedText	Args((Text));
-
-/* Variants of textToStr and syntaxOf which work for idents, ops whether
- * qualified or unqualified.
- */
 extern	String	     identToStr		Args((Cell));
-extern	Syntax	     identSyntax	Args((Cell));
+extern	Text	     fixLitText	 	Args((Text));
+extern	Text	     concatText		Args((String,String));
 
 /* --------------------------------------------------------------------------
  * Specification of syntax (i.e. default written form of application)
@@ -73,19 +77,20 @@ extern	Syntax	     identSyntax	Args((Cell));
 #define MAX_PREC  9		       /* strongest binding operator	   */
 #define FUN_PREC  (MAX_PREC+2)	       /* binding of function symbols	   */
 #define DEF_PREC  MAX_PREC
-#define APPLIC	  00000 	       /* written applicatively 	   */
-#define LEFT_ASS  02000 	       /* left associative infix	   */
-#define RIGHT_ASS 04000 	       /* right associative infix	   */
-#define NON_ASS   06000 	       /* non associative infix 	   */
-#define DEF_ASS   NON_ASS
+#define APPLIC	  0	 	       /* written applicatively 	   */
+#define LEFT_ASS  1	 	       /* left associative infix	   */
+#define RIGHT_ASS 2	 	       /* right associative infix	   */
+#define NON_ASS   3	 	       /* non associative infix 	   */
+#define DEF_ASS   LEFT_ASS
+
+#define UMINUS_PREC  6			/* Change these settings at your   */
+#define UMINUS_ASSOC LEFT_ASS		/* own risk; they may not work!	   */
 
 #define assocOf(x)	((x)&NON_ASS)
-#define precOf(x)	((x)&(~NON_ASS))
-#define mkSyntax(a,p)	((a)|(p))
+#define precOf(x)	((x)>>2)
+#define mkSyntax(a,p)	((a)|((p)<<2))
 #define DEF_OPSYNTAX	mkSyntax(DEF_ASS,DEF_PREC)
-
-extern	Void   addSyntax  Args((Int,Text,Syntax));
-extern	Syntax syntaxOf   Args((Text));
+#define NO_SYNTAX	(-1)
 
 /* --------------------------------------------------------------------------
  * Primitive functions:
@@ -115,7 +120,7 @@ struct primInfo {
 
 extern  Void registerPrims Args((struct primInfo*)); 
 extern  Void controlFuns   Args((Int)); /* Call all control functions in   */
-                                        /* prim list.                      */
+					/* prim list.                      */
 
 /* --------------------------------------------------------------------------
  * Program code storage: for holding compiled function defns etc...
@@ -145,6 +150,8 @@ extern   Heap        heapTopSnd;
 #endif
 extern   Bool        consGC;		/* Set to FALSE to turn off gc from*/
 					/* C stack; use with extreme care! */
+extern   Int	     cellsRecovered;	/* cells recovered by last gc	   */
+
 #define fst(c)	     heapTopFst[c]
 #define snd(c)	     heapTopSnd[c]
 #if PROFILING
@@ -170,22 +177,6 @@ extern  Void	     markWithoutMove Args((Cell));
 
 extern	Cell	     whatIs    Args((Cell));
 
-/* flatAlloc(n) generates an array of (n+1) Cells.
- *   flatLen(flatAlloc(n)) = n.
- *   For any flat array f, flatGet(f,0) = f.
- */
-extern  Heap	     flatspace;
-extern  Cell	     flatAlloc	     Args((Int,Int));
-#define flatOf(f)    snd(f)
-#define flatLen(f)   ((Int)(flatspace[flatOf(f)]))
-#if CHECK_TAGS
-extern Cell* flatGetChecked Args(( Cell, Int ));
-#define flatGet(f,i) *flatGetChecked(f,i)
-#else
-#define flatGet(f,i) flatspace[flatOf(f)+(i)+1]
-#endif
-#define flatSet(f,i,e) { Cell _tmp = e; flatGet(f,i) = _tmp; }
-
 /* --------------------------------------------------------------------------
  * Box cell tags are used as the fst element of a pair to indicate that
  * the snd element of the pair is to be treated in some special way, other
@@ -194,7 +185,7 @@ extern Cell* flatGetChecked Args(( Cell, Int ));
  * ------------------------------------------------------------------------*/
 
 #define TAGMIN	     1		  /* Box and constructor cell tag values   */
-#define BCSTAG	     20 	  /* Box=TAGMIN..BCSTAG-1		   */
+#define BCSTAG	     30 	  /* Box=TAGMIN..BCSTAG-1		   */
 #define isTag(c)     (TAGMIN<=(c) && (c)<SPECMIN) /* Tag cell values	   */
 #define isBoxTag(c)  (TAGMIN<=(c) && (c)<BCSTAG)  /* Box cell tag values   */
 #define isConTag(c)  (BCSTAG<=(c) && (c)<SPECMIN) /* Constr cell tag values*/
@@ -208,15 +199,17 @@ extern Cell* flatGetChecked Args(( Cell, Int ));
 #define CONIDCELL    7		  /* Identifier constructor:  snd :: Text  */
 #define CONOPCELL    8		  /* Operator constructor:    snd :: Text  */
 #define STRCELL      9		  /* String literal:	      snd :: Text  */
-#define INTCELL	     10		  /* Integer literal:	      snd :: Int   */
+#define INTCELL	     10		  /* Int literal:	      snd :: Int   */
 #if NPLUSK
 #define ADDPAT	     11		  /* (_+k) pattern discr:     snd :: Int   */
 #endif
 #if IO_HANDLES
 #define HANDCELL     12		  /* IO monad Handle:	      snd :: Int   */
 #endif
-#define DICTCELL     13		  /* Dictionary value:        snd :: Flat  */
-#define FLATCELL     14		  /* Marked flat resource:    snd :: Tag   */
+#if GC_WEAKPTRS
+#define WEAKCELL     13		  /* WeakPtr Cell:	      snd::(Cell,List)*/
+#define WEAKFIN	     14		  /* Weak pointer with fin:   snd::(k,(v,f))*/
+#endif
 #if !BREAK_FLOATS
 #define FLOATCELL    15		  /* Floating pt number:      snd :: Float */
 #endif
@@ -226,6 +219,13 @@ extern Cell* flatGetChecked Args(( Cell, Int ));
 #if GC_MALLOCPTRS
 #define MPCELL       17		  /* Malloc Ptr Cell:         snd :: Int   */
 #endif
+#if SIZEOF_INTP == SIZEOF_INT
+#define PTRCELL      18           /* C Heap Pointer:          snd ::Pointer*/
+#endif
+#if IPARAM
+#define IPCELL       19		  /* Imp Param Cell:	      snd :: Text  */
+#define IPVAR	     20		  /* ?x:		      snd :: Text  */
+#endif
 
 #define textOf(c)	((Text)(snd(c)))         /* c ::  (VAR|CON)(ID|OP) */
 #define qmodOf(c)       (textOf(fst(snd(c))))    /* c ::  QUALIDENT        */
@@ -243,6 +243,15 @@ extern Cell* flatGetChecked Args(( Cell, Int ));
 #define mkDictVar(t)	ap(DICTVAR,t)
 #define inventDictVar() mkDictVar(inventDictText())
 #define mkStr(t)	ap(STRCELL,t)
+#if IPARAM
+#define mkIParam(c)	ap(IPCELL,snd(c))
+#define isIP(p)		(whatIs(p) == IPCELL)
+#define ipMatch(pi, t)	(isIP(fun(pi)) && textOf(fun(pi)) == t)
+#define ipVar(pi)	textOf(fun(pi))
+#else
+#define isIP(p)		FALSE
+#endif
+
 extern	Bool		isVar       Args((Cell));
 extern	Bool		isCon       Args((Cell));
 extern	Bool		isQVar      Args((Cell));
@@ -261,6 +270,10 @@ extern  Cell		part2Float	Args((FloatPro));
 extern  FloatPro	floatFromParts	Args((Cell,Cell));
 #endif
 
+#define isPtr(c)        (isPair(c) && fst(c)==PTRCELL)
+extern  Cell            mkPtr           Args((Pointer));
+extern  Pointer         ptrOf           Args((Cell));
+
 /* --------------------------------------------------------------------------
  * Constructor cell tags are used as the fst element of a pair to indicate
  * a particular syntactic construct described by the snd element of the
@@ -270,73 +283,95 @@ extern  FloatPro	floatFromParts	Args((Cell,Cell));
  * element is a special cell will be treated as an application node.
  * ------------------------------------------------------------------------*/
 
-#define LETREC	     20 	  /* LETREC	snd :: ([Decl],Exp)	   */
-#define COND	     21 	  /* COND	snd :: (Exp,Exp,Exp)	   */
-#define LAMBDA	     22 	  /* LAMBDA	snd :: Alt		   */
-#define FINLIST      23 	  /* FINLIST	snd :: [Exp]		   */
-#define DOCOMP       24 	  /* DOCOMP	snd :: (Exp,[Qual])	   */
-#define BANG	     25		  /* BANG	snd :: Type		   */
-#define COMP         26 	  /* COMP	snd :: (Exp,[Qual])	   */
-#define ASPAT	     27 	  /* ASPAT	snd :: (Var,Exp)	   */
-#define ESIGN	     28 	  /* ESIGN	snd :: (Exp,Type)	   */
-#define CASE	     29 	  /* CASE	snd :: (Exp,[Alt])	   */
-#define NUMCASE	     30		  /* NUMCASE	snd :: (Exp,Disc,Rhs)	   */
-#define FATBAR	     31 	  /* FATBAR	snd :: (Exp,Exp)	   */
-#define LAZYPAT      32 	  /* LAZYPAT	snd :: Exp		   */
-#define QUAL	     33 	  /* QUAL       snd :: ([Classes],Type)    */
-#define RUNST	     34		  /* RUNST	snd :: Exp		   */
-#define DERIVE	     35		  /* DERIVE	snd :: Cell		   */
+#define LETREC	     30 	  /* LETREC	snd :: ([Decl],Exp)	   */
+#define COND	     31 	  /* COND	snd :: (Exp,Exp,Exp)	   */
+#define LAMBDA	     32 	  /* LAMBDA	snd :: Alt		   */
+#define FINLIST      33 	  /* FINLIST	snd :: [Exp]		   */
+#define DOCOMP       34 	  /* DOCOMP	snd :: (Exp,[Qual])	   */
+
+#if MUDO
+#define MDOCOMP	     44		  /* MDOCOMP	snd :: (Exp,[Qual])	   */
+#endif 
+
+#define BANG	     35		  /* BANG	snd :: Type		   */
+#define COMP         36 	  /* COMP	snd :: (Exp,[Qual])	   */
+#define ASPAT	     37 	  /* ASPAT	snd :: (Var,Exp)	   */
+#define ESIGN	     38 	  /* ESIGN	snd :: (Exp,Type)	   */
+#define RSIGN	     39 	  /* RSIGN	snd :: (Rhs,Type)	   */
+#define CASE	     40 	  /* CASE	snd :: (Exp,[Alt])	   */
+#define NUMCASE	     41		  /* NUMCASE	snd :: (Exp,Disc,Rhs)	   */
+#define FATBAR	     42 	  /* FATBAR	snd :: (Exp,Exp)	   */
+#define LAZYPAT      43 	  /* LAZYPAT	snd :: Exp		   */
+#define DERIVE	     45		  /* DERIVE	snd :: Cell		   */
 #if BREAK_FLOATS
-#define FLOATCELL    36		  /* FLOATCELL  snd :: (Int,Int)	   */
+#define FLOATCELL    46		  /* FLOATCELL  snd :: (Int,Int)	   */
 #endif
 
 #if BIGNUMS
-#define POSNUM	     37		  /* POSNUM	snd :: [Int]		   */
-#define NEGNUM	     38		  /* NEGNUM	snd :: [Int]		   */
+#define POSNUM	     47		  /* POSNUM	snd :: [Int]		   */
+#define NEGNUM	     48		  /* NEGNUM	snd :: [Int]		   */
 #endif
 
-#define BOOLQUAL     39 	  /* BOOLQUAL	snd :: Exp		   */
-#define QWHERE	     40 	  /* QWHERE	snd :: [Decl]		   */
-#define FROMQUAL     41 	  /* FROMQUAL	snd :: (Exp,Exp)	   */
-#define DOQUAL	     42		  /* DOQUAL	snd :: Exp		   */
-#define MONADCOMP    43 	  /* MONADCOMP	snd :: ((m,m0),(Exp,[Qual])*/
+#define BOOLQUAL     49 	  /* BOOLQUAL	snd :: Exp		   */
+#define QWHERE	     50 	  /* QWHERE	snd :: [Decl]		   */
+#define FROMQUAL     51 	  /* FROMQUAL	snd :: (Exp,Exp)	   */
+#define DOQUAL	     52		  /* DOQUAL	snd :: Exp		   */
+#define MONADCOMP    53 	  /* MONADCOMP	snd :: ((m,m0),(Exp,[Qual])*/
 
-#define GUARDED      44 	  /* GUARDED	snd :: [guarded exprs]	   */
+#define GUARDED      54 	  /* GUARDED	snd :: [guarded exprs]	   */
 
-#define ARRAY        45		  /* Array:     snd :: (Bounds,[Values])   */
-#define MUTVAR	     46		  /* Mutvar:	snd :: Cell		   */
+#define ARRAY        55		  /* Array	snd :: (Bounds,[Values])   */
+#define MUTVAR	     56		  /* Mutvar	snd :: Cell		   */
 #if INTERNAL_PRIMS
-#define HUGSOBJECT   47		  /* HUGSOBJECT	snd :: Cell	           */
+#define HUGSOBJECT   57		  /* HUGSOBJECT	snd :: Cell	           */
 #endif
 
-#define POLYTYPE     50		  /* POLYTYPE	snd :: (Kind,Type)	   */
-#define POLYREC	     54		  /* POLYREC:	snd :: (Int,Type)	   */
+#if STABLE_NAMES
+#define STABLENAME   58		  /* Stable Nm  snd :: Cell		   */
+#endif
 
-#define LABC	     60		  /* LABC:	snd :: (con,[(Vars,Type)]) */
-#define CONFLDS	     61		  /* CONFLDS:   snd :: (con,[Field])	   */
-#define UPDFLDS	     62		  /* UPDFLDS:   snd :: (Exp,[con],[Field]) */
+#if IPARAM
+#define WITHEXP      59 	  /* WITH	snd :: [(Var,Exp)]	   */
+#endif
+
+#define POLYTYPE     60		  /* POLYTYPE	snd :: (Kind,Type)	   */
+#define QUAL	     61		  /* QUAL	snd :: ([Classes],Type)	   */
+#define RANK2	     62		  /* RANK2	snd :: (Int,Type)	   */
+#define EXIST	     63		  /* EXIST	snd :: (Int,Type)	   */
+#define POLYREC	     64		  /* POLYREC	snd :: (Int,Type)	   */
+#define BIGLAM	     65		  /* BIGLAM	snd :: (vars,patterns)	   */
+#define CDICTS	     66		  /* CDICTS	snd :: ([Pred],Type)	   */
+
+#define LABC	     70		  /* LABC	snd :: (con,[(Vars,Type)]) */
+#define CONFLDS	     71		  /* CONFLDS	snd :: (con,[Field])	   */
+#define UPDFLDS	     72		  /* UPDFLDS	snd :: (Exp,[con],[Field]) */
 #if TREX
-#define RECORD	     63		  /* RECORD:    snd :: [Val]		   */
-#define EXTCASE	     64		  /* EXTCASE:	snd :: (Exp,Disc,Rhs)	   */
-#define RECSEL	     65		  /* RECSEL:	snd :: Ext		   */
-#define EVSUCC	     66		  /* EVSUCC:	snd :: Evidence		   */
+#define RECORD	     73		  /* RECORD	snd :: [Val]		   */
+#define EXTCASE	     74		  /* EXTCASE	snd :: (Exp,Disc,Rhs)	   */
+#define RECSEL	     75		  /* RECSEL	snd :: Ext		   */
+#endif
+#define IMPDEPS	     78		  /* IMPDEPS	snd :: [Binding]	   */
+
+#define QUALIDENT    80           /* Qualified identifier  snd :: (Id,Id)  */
+#define HIDDEN       81           /* hiding import list    snd :: [Entity] */
+#define MODULEENT    82           /* module in export list snd :: con      */
+
+#if OBSERVATIONS
+#define OBSERVEHEAD  83           /* obs. list; snd ::(first,last)         */
+#define OBSERVE      84           /* observe marker; snd :: (Cell,observe) */
+#define OBSERVESTK   85           /* observe marker on stack               */
 #endif
 
-#define QUALIDENT    70           /* Qualified identifier  snd :: (Id,Id)  */
-#define HIDDEN       71           /* hiding import list    snd :: [Entity] */
-#define MODULEENT    72           /* module in export list snd :: con      */
+#define INFIX	     90		  /* INFIX	snd :: (see tidyInfix)	   */
+#define ONLY	     91		  /* ONLY	snd :: Exp		   */
+#define NEG	     92		  /* NEG	snd :: Exp		   */
 
-#define ONLY	     80		  /* ONLY:      snd :: Exp (used in parser)*/
-#define NEG	     81		  /* NEG:       snd :: Exp (used in parser)*/
+#if ZIP_COMP
+#define ZCOMP        93 	  /* ZCOMP	snd :: (Exp,[[Qual]])	   */
+#endif
 
-#if GENERATE_STG
-#define STGVAR       90           /* STGVAR     snd :: (StgRhs,info)       */
-#define STGAPP       91           /* STGAPP     snd :: (StgVar,[Arg])      */
-#define STGPRIM      92           /* STGPRIM    snd :: (PrimOp,[Arg])      */
-#define STGCON       93           /* STGCON     snd :: (StgCon,[Arg])      */
-#define PRIMCASE     94           /* PRIMCASE   snd :: (Expr,([Var],Expr)) */
-#define LETALT       95           /* LETALT     snd :: (AltVar,Expr)       */
-#define STGALTVAR    96           /* STGALTVAR  snd :: (Expr,info)         */
+#if SIZEOF_INTP != SIZEOF_INT
+#define PTRCELL      99           /* C Heap Pointer snd :: (Int,Int)       */
 #endif
 
 /* Last constructor tag must be less than SPECMIN */
@@ -355,7 +390,9 @@ extern  FloatPro	floatFromParts	Args((Cell,Cell));
 #endif
 #define WILDCARD     104	  /* Wildcard pattern			   */
 
-#define DOTDOT       105          /* ".." in import/export list            */
+#define SKOLEM	     105	  /* Skolem constant			   */
+
+#define DOTDOT       106          /* ".." in import/export list            */
 
 #if BIGNUMS
 #define ZERONUM      108	  /* The zero bignum (see POSNUM, NEGNUM)  */
@@ -364,7 +401,7 @@ extern  FloatPro	floatFromParts	Args((Cell,Cell));
 #define NAME	     110	  /* whatIs code for isName		   */
 #define TYCON	     111	  /* whatIs code for isTycon		   */
 #define CLASS	     112	  /* whatIs code for isClass		   */
-#define SELECT       113          /* whatIs code for isSelect              */
+#define MODULE       113          /* whatIs code for isModule              */
 #define INSTANCE     114          /* whatIs code for isInst                */
 #define TUPLE	     115	  /* whatIs code for tuple constructor	   */
 #define OFFSET	     116	  /* whatis code for offset		   */
@@ -375,16 +412,17 @@ extern  FloatPro	floatFromParts	Args((Cell,Cell));
 #endif
 
 #define SIGDECL      120	  /* Signature declaration		   */
-#define PREDEFINED   121	  /* predefined name, not yet filled       */
+#define FIXDECL      121	  /* Fixity declaration			   */
+#define FUNBIND	     122	  /* Function binding			   */
+#define PATBIND	     123	  /* Pattern binding			   */
 
-#define DATATYPE     130	  /* datatype type constructor		   */
-#define NEWTYPE	     131	  /* newtype type constructor		   */
-#define SYNONYM	     132	  /* synonym type constructor		   */
-#define RESTRICTSYN  133	  /* synonym with restricted scope	   */
+#define DATATYPE     130	  /* Datatype type constructor		   */
+#define NEWTYPE	     131	  /* Newtype type constructor		   */
+#define SYNONYM	     132	  /* Synonym type constructor		   */
+#define RESTRICTSYN  133	  /* Synonym with restricted scope	   */
 
-#define MODULE       134          /* whatIs code for isModule              */
-
-#define NODEPENDS    135	  /* stop calculation of deps in type check*/
+#define NODEPENDS    135	  /* Stop calculation of deps in type check*/
+#define PREDEFINED   136	  /* Predefined name, not yet filled       */
 
 /* --------------------------------------------------------------------------
  * Tuple data/type constructors:
@@ -425,11 +463,64 @@ extern Ext           mkExt Args((Text));
 #define offsetOf(c)  ((c)-OFFMIN)
 #define mkOffset(o)  (OFFMIN+(o))
 
+/* --------------------------------------------------------------------------
+ * Observations
+ * ------------------------------------------------------------------------*/
+
+#if OBSERVATIONS
+#define OBSMIN	      (OFFMIN+NUM_OFFSETS)
+#define observe(n)    tabObserve[(n)-OBSMIN]
+
+struct Observe {
+	Text tag;		/* tag in observe primitive		*/
+	Cell head;
+};
+
+#define firstObs(c) snd3(c)
+#define lastObs(c)  thd3(c)
+
+#define nextObs(c)  fst3(c)
+#define seqObs(c)   snd3(c)
+#define exprObs(c)  thd3(c)
+
+#define markedExpr(c) snd3(c)
+#define markedObs(c)  thd3(c)
+
+extern struct    Observe          DECTABLE(tabObserve);
+extern Observe   firstObserve	  Args((Void));
+extern Observe   nextObserve	  Args((Void));
+extern Void      clearObserve	  Args((Void));
+extern Cell      addObsInstance	  Args((String,Cell,Int));
+extern Void      insertAfterObs   Args((Cell,Cell));
+
+/* --------------------------------------------------------------------------
+ * Breakpoints
+ * ------------------------------------------------------------------------*/
+
+#define BRKMIN	      (OBSMIN+NUM_OBS_TAGS)
+#define breakpt(n)    tabBreakpt[(n)-BRKMIN]
+
+struct Breakpt {
+	Text tag;
+	Bool enabled;
+	Int  count;			/* number of breaks to skip 	   */
+};
+
+extern Void     clearAllBreak   Args((Void));
+extern Bool     breakNow	Args((String));
+extern Void     setBreakpt	Args((String,Bool)); 
+extern Void     setBreakCount   Args((String,Int));
+#endif
+
 /* --------------------------------------------------------------------------
  * Modules:
  * ------------------------------------------------------------------------*/
 
-#define MODMIN	      (OFFMIN+NUM_OFFSETS)
+#if OBSERVATIONS
+#define MODMIN        (BRKMIN+NUM_BRKPTS)
+#else
+#define MODMIN        (OFFMIN+NUM_OFFSETS)
+#endif
 
 #if IGNORE_MODULES
 #define setCurrModule(m) doNothing()
@@ -501,20 +592,22 @@ extern Tycon newTycon	  Args((Text));
 extern Tycon findTycon	  Args((Text));
 extern Tycon addTycon	  Args((Tycon));
 extern Tycon findQualTycon Args((Cell));
-extern Tycon addPrimTycon Args((String,Kind,Int,Cell,Cell));
+extern Tycon addPrimTycon Args((Text,Kind,Int,Cell,Cell));
 
 #define isSynonym(h)	(isTycon(h) && tycon(h).what==SYNONYM)
+#define isQualType(t)	(isPair(t) && fst(t)==QUAL)
 #define mkPolyType(n,t)	pair(POLYTYPE,pair(n,t))
 #define isPolyType(t)	(isPair(t) && fst(t)==POLYTYPE)
+#define isPolyOrQualType(t) (isPair(t) && (fst(t)==POLYTYPE || fst(t)==QUAL))
 #define polySigOf(t)	fst(snd(t))
-#define monoTypeOf(t)	snd(snd(t))
+#define monotypeOf(t)	snd(snd(t))
 
 /* --------------------------------------------------------------------------
  * Globally defined name values:
  * ------------------------------------------------------------------------*/
 
 #define NAMEMIN      (TYCMIN+NUM_TYCON)
-#define isName(c)    (NAMEMIN<=(c) && (c)<SELMIN)
+#define isName(c)    (NAMEMIN<=(c) && (c)<INSTMIN)
 #define mkName(n)    (NAMEMIN+(n))
 #define name(n)      tabName[(n)-NAMEMIN]
 
@@ -524,18 +617,16 @@ struct strName {
 #if !IGNORE_MODULES
     Module mod;                         /* module that defines it          */
 #endif
+    Syntax syntax;
     Cell parent;                        /* enclosing object                */
-                                        /*   :: (Class|Instance,Name)      */
-                                        /*      | Name                     */
-                                        /*      | NIL                      */
+					/* :: Class | Instance | Name | NIL*/
     Int  arity;
     Int  number;
     Cell type;
     Cell defn;
-#if GENERATE_STG 
-    Cell stgVar;
-#endif
     Addr code;
+    Text callconv;                      /* for foreign import/export       */
+    Text extFun;                        /* for foreign import/export       */
     Prim primDef;
     Name nextNameHash;
 #if PROFILING
@@ -547,8 +638,11 @@ extern struct strName DECTABLE(tabName);
 
 /* The number field in a name is used to distinguish various kinds of name:
  *   mfunNo(i) = code for member function, offset i
+ *		 members that are sole elements of dict use mfunNo(0)
+ *		 members of dicts with more than one elem use mfunNo(n), n>=1
  *   EXECNAME  = code for executable name (bytecodes or primitive)
  *   SELNAME   = code for selector function
+ *   DFUNNAME  = code for dictionary builder or selector
  *   cfunNo(i) = code for data constructor
  *		 datatypes with only one constructor uses cfunNo(0)
  *		 datatypes with multiple constructors use cfunNo(n), n>=1
@@ -556,24 +650,27 @@ extern struct strName DECTABLE(tabName);
 
 #define EXECNAME	0
 #define SELNAME		1
-#define MIN_CFUN        2
+#define DFUNNAME	2
+#define CFUNNAME	3
 
 #define isSfun(n)	(name(n).number==SELNAME)
+#define isDfun(n)	(name(n).number==DFUNNAME)
 
-#define isCfun(n)	(name(n).number>=MIN_CFUN)
-#define cfunOf(n)	(name(n).number-MIN_CFUN)
-#define cfunNo(i)	((i)+MIN_CFUN)
+#define isCfun(n)	(name(n).number>=CFUNNAME)
+#define cfunOf(n)	(name(n).number-CFUNNAME)
+#define cfunNo(i)	((i)+CFUNNAME)
 #define hasCfun(cs)	(nonNull(cs) && isCfun(hd(cs)))
 
 #define isMfun(n)	(name(n).number<0)
-#define mfunOf(n)	(-name(n).number)
-#define mfunNo(i)	(-(i))
+#define mfunOf(n)	((-1)-name(n).number)
+#define mfunNo(i)	((-1)-(i))
 
 extern Name   newName	   Args((Text,Cell));
 extern Name   findName	   Args((Text));
 extern Name   addName	   Args((Name));
-extern Name   findQualName Args((Int,Cell));
-extern Void   addPrim	   Args((Int,Name,String,Type));
+extern Name   findQualName Args((Cell));
+extern Name   findQualFun  Args((Text,Text));
+extern Void   addPrim	   Args((Int,Name,String,Module,Type));
 extern Name   addPrimCfun  Args((Text,Int,Int,Cell));
 extern Int    sfunPos	   Args((Name,Name));
 
@@ -581,12 +678,7 @@ extern Int    sfunPos	   Args((Name,Name));
  * Type class values:
  * ------------------------------------------------------------------------*/
 
-#define SELMIN       (NAMEMIN+NUM_NAME)          /* dictionary selectors   */
-#define isSelect(c)  (SELMIN<=(c) && (c)<INSTMIN)
-#define mkSelect(n)  (SELMIN+(n))
-#define selectOf(c)  ((Int)((c)-SELMIN))
-
-#define INSTMIN      (SELMIN+NUM_SELECTS)        /* instances              */
+#define INSTMIN      (NAMEMIN+NUM_NAME)	/* instances			   */
 #define isInst(c)    (INSTMIN<=(c) && (c)<CLASSMIN)
 #define mkInst(n)    (INSTMIN+(n))
 #define instOf(c)    ((Int)((c)-INSTMIN))
@@ -594,14 +686,13 @@ extern Int    sfunPos	   Args((Name,Name));
 
 struct strInst {
     Class c;				/* class C			   */
-    Cell  t;				/* type  T			   */
-    Int   arity;			/* number of args		   */
     Int   line;
+    Kinds kinds;			/* Kinds of variables in head	   */
+    Cell  head;				/* :: Pred			   */
     List  specifics;			/* :: [Pred]			   */
     Int   numSpecifics;			/* length(specifics)		   */
     List  implements;
-    List  dicts;			/* :: [Dict]			   */
-    List  superBuild;			/* instructions for superclasses   */
+    Name  builder;			/* Dictionary constructor function */
 };
 
 /* a predicate (an element :: Pred) is an application of a Class to one or
@@ -614,58 +705,38 @@ struct strInst {
 #define cclass(n)    tabClass[(n)-CLASSMIN]
 
 struct strClass {
-    Text text;				/* Name of class		   */
-    Int  line;				/* Line where declaration begins   */
+    Text   text;			/* Name of class		   */
+    Int    line;			/* Line where declaration begins   */
 #if !IGNORE_MODULES
-    Module mod;                         /* module that declares it         */
+    Module mod;				/* module that declares it         */
 #endif
-    Int  level;				/* Level in class hierarchy	   */
-    Kind sig;				/* Kind of constructors in class   */
-    List supers;			/* :: [Class] (immed superclasses) */
-    Int  numSupers;			/* length(supers)		   */
-    List members;			/* :: [Name]			   */
-    Int  numMembers;			/* length(members)		   */
-    List defaults;			/* :: [Name]			   */
-    List instances;			/* :: [Inst]			   */
+    Int    level;			/* Level in class hierarchy	   */
+    Int    arity;			/* Number of arguments		   */
+    Kinds  kinds;			/* Kinds of constructors in class  */
+    Cell   head;			/* Head of class		   */
+    List   fds;				/* Functional Dependencies	   */
+    List   xfds;			/* Xpanded Functional Dependencies */
+    Name   dcon;			/* Dictionary constructor function */
+    List   supers;			/* :: [Pred]			   */
+    Int    numSupers;			/* length(supers)		   */
+    List   dsels;			/* Superclass dictionary selectors */
+    List   members;			/* :: [Name]			   */
+    Int    numMembers;			/* length(members)		   */
+    List   defaults;			/* :: [Name]			   */
+    List   instances;			/* :: [Inst]			   */
 };
 
 extern struct strClass    DECTABLE(tabClass);
 extern struct strInst far *tabInst;
 
-#if CHECK_TAGS
-#define checkDict(c) \
-  if (fst(c) != DICTCELL) internal("checkDict1"); \
-  if (flatGet(c,0) != c)  internal("checkDict2"); 
-extern Int  dictOf  Args((Cell));
-extern Cell dictGet Args((Cell,Int));
-extern Void dictSet Args((Cell,Int,Cell));
-#else
-#define dictOf(c)		flatOf(c)
-#define dictGet(d,i)		flatGet(d,i)
-#define dictSet(d,i,e)		flatSet(d,i,e)
-#endif /* CHECK_TAGS */
-extern Cell newDict             Args((Int));
-
-#define dictMembersStart(cl)	(1)
-#define dictSupersStart(cl)	(dictMembersStart(cl) + cclass(cl).numMembers)
-#define dictSpecificsStart(cl)	(dictSupersStart(cl) + cclass(cl).numSupers)
-#define dictLength(in)		(dictSpecificsStart(inst(in).c) + \
-				 inst(in).numSpecifics)
-
-#if GENERATE_STG
-extern Inst  instHw;            /* next unused instance record      */
-#endif
-
 extern Class newClass	   Args((Text));
 extern Class classMax	   Args((Void));
 extern Class findClass	   Args((Text));
 extern Class addClass	   Args((Class));
 extern Class findQualClass Args((Cell));
 extern Inst  newInst	   Args((Void));
-extern Inst  findInst	   Args((Class,Type));
 extern Inst  findFirstInst Args((Tycon));
 extern Inst  findNextInst  Args((Tycon,Inst));
-extern Cell  makeInstPred  Args((Inst));
 
 /* --------------------------------------------------------------------------
  * Character values:
@@ -675,16 +746,18 @@ extern Cell  makeInstPred  Args((Inst));
 #define MAXCHARVAL   (NUM_CHARS-1)
 #define isChar(c)    (CHARMIN<=(c) && (c)<INTMIN)
 #define charOf(c)    ((Char)(c-CHARMIN))
-#define mkChar(c)    ((Cell)(CHARMIN+((unsigned)((c)%NUM_CHARS))))
+#define mkChar(c)    ((Cell)(CHARMIN+(((unsigned)(c))%NUM_CHARS)))
 
 /* --------------------------------------------------------------------------
  * Small Integer values:
  * ------------------------------------------------------------------------*/
 
 #define INTMIN	     (CHARMIN+NUM_CHARS)
-#define INTMAX	     MAXPOSINT
+#define INTMAX	     (MAXPOSINT)
 #define isSmall(c)   (INTMIN<=(c))
 #define INTZERO      (INTMIN/2 + INTMAX/2)
+#define MINSMALLINT  (INTMIN - INTZERO)
+#define MAXSMALLINT  (INTMAX - INTZERO)
 #define mkDigit(c)   ((Cell)((c)+INTMIN))
 #define digitOf(c)   ((Int)((c)-INTMIN))
 
@@ -718,30 +791,27 @@ extern  Bool isBignum Args((Cell));
 #define tl(c)	     snd(c)
 
 extern	Int	     length	  Args((List));
-extern	List	     appendOnto   Args((List,List));
+extern	List	     appendOnto   Args((List,List));	/* destructive	   */
 extern	List	     dupOnto      Args((List,List));
 extern  List	     dupList	  Args((List));
-extern	List	     revOnto	  Args((List, List));
-#define rev(xs)      revOnto((xs),NIL)
-#if 0
-extern  List	     minus	  Args((List,List));
-#endif
+extern  List	     dupUpto	  Args((List,Cell));
+extern	List	     revOnto	  Args((List, List));	/* destructive	   */
+#define rev(xs)      revOnto((xs),NIL)			/* destructive	   */
 extern	Cell	     cellIsMember Args((Cell,List));
 extern  Cell         cellAssoc    Args((Cell,List));
 extern  Cell         cellRevAssoc Args((Cell,List));
+extern	List	     concat	  Args((List));
+extern	List	     intersect	  Args((List,List));
 extern	Cell	     varIsMember  Args((Text,List));
+extern	Name	     nameIsMember Args((Text,List));
 extern  Cell	     intIsMember  Args((Int,List));
-extern	List	     copy	  Args((Int,Cell));
-extern	List	     diffList	  Args((List,List));
-extern  List         deleteCell   Args((List,Cell));
-extern  List	     take	  Args((Int,List));
-extern  List	     splitAt	  Args((Int,List));
-#if 0 /* Not used in this version of Hugs */
-extern  Bool         eqList       Args((List,List));
-#endif
+extern	List	     replicate	  Args((Int,Cell));
+extern	List	     diffList	  Args((List,List));	/* destructive	   */
+extern  List         deleteCell   Args((List,Cell));	/* non-destructive */
+extern  List	     take	  Args((Int,List));	/* destructive	   */
+extern  List	     splitAt	  Args((Int,List));	/* non-destructive */
 extern  Cell	     nth	  Args((Int,List));
-extern  List	     initSeg	  Args((List));
-extern  List	     removeCell	  Args((Cell,List));
+extern  List	     removeCell	  Args((Cell,List));	/* destructive	   */
 
 /* The following macros provide `inline expansion' of some common ways of
  * traversing, using and modifying lists:
@@ -750,20 +820,22 @@ extern  List	     removeCell	  Args((Cell,List));
  *	with identifiers used elsewhere.
  */
 
-#define mapBasic(_init,_step)	  {List Zs=(_init);\
-				   for(;nonNull(Zs);Zs=tl(Zs))  \
-				   _step;}
-#define mapModify(_init,_step)	  mapBasic(_init,hd(Zs)=_step)
+#define mapBasic(_init,_step)		{List Zs=(_init);\
+					 for(;nonNull(Zs);Zs=tl(Zs))  \
+					 _step;}
+#define mapModify(_init,_step)		mapBasic(_init,hd(Zs)=_step)
 
-#define mapProc(_f,_xs) 	  mapBasic(_xs,_f(hd(Zs)))
-#define map1Proc(_f,_a,_xs)	  mapBasic(_xs,_f(_a,hd(Zs)))
-#define map2Proc(_f,_a,_b,_xs)	  mapBasic(_xs,_f(_a,_b,hd(Zs)))
-#define map3Proc(_f,_a,_b,_c,_xs) mapBasic(_xs,_f(_a,_b,_c,hd(Zs)))
+#define mapProc(_f,_xs)			mapBasic(_xs,_f(hd(Zs)))
+#define map1Proc(_f,_a,_xs)		mapBasic(_xs,_f(_a,hd(Zs)))
+#define map2Proc(_f,_a,_b,_xs)		mapBasic(_xs,_f(_a,_b,hd(Zs)))
+#define map3Proc(_f,_a,_b,_c,_xs)	mapBasic(_xs,_f(_a,_b,_c,hd(Zs)))
+#define map4Proc(_f,_a,_b,_c,_d,_xs)	mapBasic(_xs,_f(_a,_b,_c,_d,hd(Zs)))
 
-#define mapOver(_f,_xs) 	  mapModify(_xs,_f(hd(Zs)))
-#define map1Over(_f,_a,_xs)	  mapModify(_xs,_f(_a,hd(Zs)))
-#define map2Over(_f,_a,_b,_xs)	  mapModify(_xs,_f(_a,_b,hd(Zs)))
-#define map3Over(_f,_a,_b,_c,_xs) mapModify(_xs,_f(_a,_b,_c,hd(Zs)))
+#define mapOver(_f,_xs)			mapModify(_xs,_f(hd(Zs)))
+#define map1Over(_f,_a,_xs)		mapModify(_xs,_f(_a,hd(Zs)))
+#define map2Over(_f,_a,_b,_xs)		mapModify(_xs,_f(_a,_b,hd(Zs)))
+#define map3Over(_f,_a,_b,_c,_xs)	mapModify(_xs,_f(_a,_b,_c,hd(Zs)))
+#define map4Over(_f,_a,_b,_c,_d,_xs)	mapModify(_xs,_f(_a,_b,_c,_d,hd(Zs)))
 
 /* This is just what you want for functions with accumulating parameters */
 #define mapAccum(_f,_acc,_xs) 	        mapBasic(_xs,_acc=_f(_acc,hd(Zs)))
@@ -776,6 +848,9 @@ extern  List	     removeCell	  Args((Cell,List));
  * ------------------------------------------------------------------------*/
 
 #define ap(f,x)      pair(f,x)
+#define ap1(f,x)     ap(f,x)
+#define ap2(f,x,y)   ap(ap(f,x),y)
+#define ap3(f,x,y,z) ap(ap(ap(f,x),y),z)
 #define fun(c)	     fst(c)
 #define arg(c)	     snd(c)
 #define isAp(c)      (isPair(c) && !isTag(fst(c)))
@@ -803,7 +878,7 @@ extern	StackPtr sp;
 #define clearStack() sp=(-1)
 #define stackEmpty() (sp==(-1))
 #define stack(p)     cellStack[p]
-#define chkStack(n)  if (sp>=NUM_STACK-(n)) stackOverflow()
+#define chkStack(n)  if (sp>=NUM_STACK-(n)) hugsStackOverflow()
 #define push(c)      \
   do {               \
     chkStack(1);     \
@@ -817,7 +892,16 @@ extern	StackPtr sp;
 #define topfun(f)    top()=ap((f),top())
 #define toparg(x)    top()=ap(top(),(x))
 
-extern  Void stackOverflow Args((Void));
+extern  Void hugsStackOverflow Args((Void));
+
+#if __MWERKS__ && macintosh
+#include <Memory.h>
+#define STACK_HEADROOM 16384
+#define STACK_CHECK if (StackSpace() <= STACK_HEADROOM) \
+		      internal("Macintosh function parameter stack overflow.");
+#else
+#define STACK_CHECK
+#endif
 
 /* --------------------------------------------------------------------------
  * Script file control:
@@ -862,11 +946,11 @@ extern Cell   openHandle Args((String,Int,Bool));
 extern struct strHandle  DECTABLE(handles);
 #endif
 
-#if GC_MALLOCPTRS
 /* --------------------------------------------------------------------------
  * Malloc Pointers
  * ------------------------------------------------------------------------*/
 
+#if GC_MALLOCPTRS
 struct strMallocPtr {		/* Malloc Ptr description                  */
     Cell mpcell;		/* Back pointer to MPCELL                  */
     Void *ptr;			/* Pointer into C world                    */
@@ -881,9 +965,22 @@ extern Void   incMallocPtrRefCnt Args((Int, Int));
 
 #define mpOf(c)	   snd(c)
 #define derefMP(c) (mallocPtrs[(Int)mpOf(c)].ptr)
-#define sizeofMallocPtr 2
 #endif /* GC_MALLOCPTRS */
 
+#if GC_WEAKPTRS
+/* --------------------------------------------------------------------------
+ * Weak Pointers
+ * ------------------------------------------------------------------------*/
+
+#define mkWeakPtr(c)    pair(WEAKCELL,pair(c,NIL))
+#define derefWeakPtr(c) fst(snd(c))
+#define nextWeakPtr(c) snd(snd(c))
+
+extern List finalizers;
+extern List liveWeakPtrs;
+
+#endif /* GC_WEAKPTRS */
+
 /* --------------------------------------------------------------------------
  * Stable pointers
  * ------------------------------------------------------------------------*/
@@ -900,9 +997,9 @@ extern  Void freeStablePtr   Args((Int));
 
 /* This is an exact copy of the declaration found in GreenCard.h */
 
-typedef int   HugsStackPtr;
-typedef int   HugsStablePtr;
-typedef void* HugsForeign;
+typedef int     HugsStackPtr;
+typedef int     HugsStablePtr;
+typedef Pointer HugsForeign;
 
 typedef struct {
 
@@ -919,7 +1016,7 @@ typedef struct {
   /* push part of result   */
   void      	 (*putInt   )     Args((int));           
   void      	 (*putWord  )     Args((unsigned int));
-  void      	 (*putAddr  )     Args((void *));
+  void      	 (*putAddr  )     Args((void*));
   void      	 (*putFloat )     Args((double));
   void      	 (*putDouble)     Args((double));
   void      	 (*putChar  )     Args((char));
@@ -937,6 +1034,59 @@ typedef struct {
   /* register the prim table */	    			 
   void      	 (*registerPrims) Args((struct primInfo*));
 			   
+  /* garbage collect */
+  void		 (*garbageCollect) Args(());
+
+  /* API3 additions follow */
+  HugsStablePtr  (*lookupName)     Args((char*, char*));
+  void           (*ap)             Args((int));
+  void           (*getUnit)        Args(());
+  void*          (*mkThunk)        Args((void*, HugsStablePtr));
+  void           (*freeThunk)      Args((void*));
+  int     	 (*getBool)        Args(());
+  void      	 (*putBool)        Args((int));
+
+} HugsAPI3;
+
+extern  HugsAPI3* hugsAPI3     Args((Void));
+typedef Void (*InitModuleFun3) Args((HugsAPI3*));
+
+typedef struct {
+
+  /* evaluate next argument */
+  int            (*getInt   )     Args(());  
+  unsigned int   (*getWord  )     Args(());
+  void*     	 (*getAddr  )     Args(());
+  float     	 (*getFloat )     Args(());
+  double    	 (*getDouble)     Args(());
+  char      	 (*getChar  )     Args(());
+  HugsForeign    (*getForeign)    Args(());
+  HugsStablePtr  (*getStablePtr)  Args(());
+
+  /* push part of result   */
+  void      	 (*putInt   )     Args((int));           
+  void      	 (*putWord  )     Args((unsigned int));
+  void      	 (*putAddr  )     Args((void*));
+  void      	 (*putFloat )     Args((double));
+  void      	 (*putDouble)     Args((double));
+  void      	 (*putChar  )     Args((char));
+  void      	 (*putForeign)    Args((HugsForeign, void (*)(HugsForeign)));
+  void      	 (*putStablePtr)  Args((HugsStablePtr));
+
+  /* return n values in IO monad or Id monad */
+  void      	 (*returnIO)      Args((HugsStackPtr, int));
+  void      	 (*returnId)      Args((HugsStackPtr, int));
+  int      	 (*runIO)         Args((int));
+
+  /* free a stable pointer */	    			 
+  void      	 (*freeStablePtr) Args((HugsStablePtr));
+
+  /* register the prim table */	    			 
+  void      	 (*registerPrims) Args((struct primInfo*));
+			   
+  /* garbage collect */
+  void		 (*garbageCollect) Args(());
+
 } HugsAPI2;
 
 extern  HugsAPI2* hugsAPI2     Args((Void));
@@ -992,4 +1142,13 @@ typedef struct {
 extern  HugsAPI1* hugsAPI1     Args((Void));
 typedef Void (*InitModuleFun1) Args((HugsAPI1*));
 
+/* --------------------------------------------------------------------------
+ * Misc:
+ * ------------------------------------------------------------------------*/
+
+extern  Void   setLastExpr       Args((Cell));
+extern  Cell   getLastExpr       Args((Void));
+extern  List   addTyconsMatching Args((String,List));
+extern  List   addNamesMatching  Args((String,List));
+
 /*-------------------------------------------------------------------------*/
